#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os
import time
import tkinter as tk

from pathlib import Path
from tkinter import messagebox
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askdirectory, asksaveasfilename

from reclaimer.hek.defs.antr import antr_def
from reclaimer.animation.jma import read_jma, write_jma,\
     JmaAnimation, JmaAnimationSet, JMA_ANIMATION_EXTENSIONS
from reclaimer.animation.animation_compilation import \
     compile_model_animations, ANIMATION_COMPILE_MODE_NEW,\
     ANIMATION_COMPILE_MODE_PRESERVE, ANIMATION_COMPILE_MODE_ADDITIVE
from reclaimer.animation.util import partial_mod2_def

from supyr_struct.util import is_in_dir, path_normalize,\
     path_split, path_replace

from mozzarilla import editor_constants as e_c

if __name__ == "__main__":
    window_base_class = tk.Tk
else:
    window_base_class = tk.Toplevel


class AnimationsCompilerWindow(window_base_class, BinillaWidget):
    app_root = None
    tags_dir = ''

    jma_anims = ()
    jma_anim_set = None

    _compiling = False
    _loading = False
    _saving = False

    _jma_tree_iids = ()

    animation_delta_tolerance = 0.00001

    def __init__(self, app_root, *args, **kwargs):
        if window_base_class == tk.Toplevel:
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
            self.app_root = app_root
        else:
            self.app_root = self

        window_base_class.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, *args, **kwargs)

        self.title("Model_animations compiler")
        self.resizable(1, 1)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        tags_dir = getattr(app_root, "tags_dir", "")

        self.tags_dir = tk.StringVar(self, tags_dir if tags_dir else "")
        self.jma_dir = tk.StringVar(self)
        self.model_animations_path = tk.StringVar(self)

        self.animation_count_limit = tk.IntVar(self, 256)
        self.calculate_limp_limb_vectors = tk.IntVar(self, 0)
        self.update_mode = tk.IntVar(self, ANIMATION_COMPILE_MODE_PRESERVE)
        self.animation_delta_tolerance_string = tk.StringVar(
            self, str(self.animation_delta_tolerance))
        self.animation_delta_tolerance_string.trace(
            "w", lambda *a, s=self: s.set_animation_delta_tolerance())


        # make the frames
        self.main_frame = tk.Frame(self)
        self.jma_info_frame = tk.LabelFrame(
            self, text="Animations info")

        self.dirs_frame = tk.LabelFrame(
            self.main_frame, text="Directories")
        self.buttons_frame = tk.Frame(self.main_frame)
        self.settings_frame = tk.Frame(self.main_frame)

        self.jma_dir_frame = tk.LabelFrame(
            self.dirs_frame, text="Source animations folder")
        self.tags_dir_frame = tk.LabelFrame(
            self.dirs_frame, text="Tags directory root folder")
        self.model_animations_path_frame = tk.LabelFrame(
            self.dirs_frame, text="Model_animations output path")

        self.animation_delta_tolerance_frame = tk.LabelFrame(
            self.settings_frame, text="Animation delta tolerance")
        self.update_mode_frame = tk.LabelFrame(
            self.settings_frame, text="What to do with existing model_animations tag")

        self.compile_mode_replace_rbtn = tk.Radiobutton(
            self.update_mode_frame, anchor="w",
            variable=self.update_mode, value=ANIMATION_COMPILE_MODE_NEW,
            text="Erase all animations/tag values")
        self.compile_mode_preserve_rbtn = tk.Radiobutton(
            self.update_mode_frame, anchor="w",
            variable=self.update_mode, value=ANIMATION_COMPILE_MODE_PRESERVE,
            text="Preserve any used animations/tag values")
        self.compile_mode_additive_rbtn = tk.Radiobutton(
            self.update_mode_frame, anchor="w",
            variable=self.update_mode, value=ANIMATION_COMPILE_MODE_ADDITIVE,
            text="Erase nothing(only add/update animations and values)")

        self.animation_delta_tolerance_info = tk.Label(
            self.animation_delta_tolerance_frame, justify='left', anchor="w",
            text=("How much a nodes position, rotation, or scale\n"
                  "must change from the starting frame for that\n"
                  "type of transform to be considered animated."))
        self.animation_delta_tolerance_spinbox = tk.Spinbox(
            self.animation_delta_tolerance_frame, from_=0,
            to=100, width=25, increment=self.animation_delta_tolerance,
            textvariable=self.animation_delta_tolerance_string, justify="right")

        self.use_os_animation_count_limit_cbtn = tk.Checkbutton(
            self.settings_frame, onvalue=2048, offvalue=256,
            variable=self.animation_count_limit, anchor="w",
            text="Use Open Sauce animation count limit")
        self.calculate_limp_limb_vectors_cbtn = tk.Checkbutton(
            self.settings_frame, variable=self.calculate_limp_limb_vectors,
            text=("Calculate biped limp body node vectors\n"
                  "(requires matching gbxmodel)"), anchor="w")


        self.jma_info_tree = tk.ttk.Treeview(
            self.jma_info_frame, selectmode='browse', padding=(0, 0), height=4)
        self.jma_info_vsb = tk.Scrollbar(
            self.jma_info_frame, orient='vertical',
            command=self.jma_info_tree.yview)
        self.jma_info_hsb = tk.Scrollbar(
            self.jma_info_frame, orient='horizontal',
            command=self.jma_info_tree.xview)
        self.jma_info_tree.config(yscrollcommand=self.jma_info_vsb.set,
                                  xscrollcommand=self.jma_info_hsb.set)

        self.jma_dir_entry = tk.Entry(
            self.jma_dir_frame, textvariable=self.jma_dir, state=tk.DISABLED)
        self.jma_dir_browse_button = tk.Button(
            self.jma_dir_frame, text="Browse", command=self.jma_dir_browse)


        self.tags_dir_entry = tk.Entry(
            self.tags_dir_frame, textvariable=self.tags_dir, state=tk.DISABLED)
        self.tags_dir_browse_button = tk.Button(
            self.tags_dir_frame, text="Browse", command=self.tags_dir_browse)


        self.model_animations_path_entry = tk.Entry(
            self.model_animations_path_frame,
            textvariable=self.model_animations_path,
            state=tk.DISABLED)
        self.model_animations_path_browse_button = tk.Button(
            self.model_animations_path_frame, text="Browse",
            command=self.model_animations_path_browse)


        self.load_button = tk.Button(
            self.buttons_frame, text="Load\nanimations",
            command=self.load_animations)
        self.save_button = tk.Button(
            self.buttons_frame, text="Save as JMA",
            command=self.save_animations)
        self.compile_button = tk.Button(
            self.buttons_frame, text="Compile\nmodel_animations",
            command=self.compile_model_animations)

        self.populate_animations_info_tree()

        # pack everything
        self.main_frame.pack(fill="both", side='left', pady=3, padx=3)
        self.jma_info_frame.pack(fill="both", side='left', pady=3, padx=3,
                                 expand=True)

        self.dirs_frame.pack(fill="x")
        self.buttons_frame.pack(fill="x", pady=3, padx=3)
        self.settings_frame.pack(fill="both")

        self.jma_dir_frame.pack(fill='x')
        self.tags_dir_frame.pack(fill='x')
        self.model_animations_path_frame.pack(fill='x')

        self.jma_dir_entry.pack(side='left', fill='x', expand=True)
        self.jma_dir_browse_button.pack(side='left')

        self.model_animations_path_entry.pack(side='left', fill='x', expand=True)
        self.model_animations_path_browse_button.pack(side='left')

        self.tags_dir_entry.pack(side='left', fill='x', expand=True)
        self.tags_dir_browse_button.pack(side='left')

        self.jma_info_hsb.pack(side="bottom", fill='x')
        self.jma_info_vsb.pack(side="right",  fill='y')
        self.jma_info_tree.pack(side='left', fill='both', expand=True)

        self.load_button.pack(side='left', fill='both', padx=3, expand=True)
        self.save_button.pack(side='left', fill='both', padx=3, expand=True)
        self.compile_button.pack(side='right', fill='both', padx=3, expand=True)

        for w in (self.update_mode_frame,
                  self.animation_delta_tolerance_frame):
            w.pack(expand=True, fill='both')

        for w in (self.compile_mode_replace_rbtn,
                  self.compile_mode_preserve_rbtn,
                  self.compile_mode_additive_rbtn,):
            w.pack(expand=True, fill='both')

        self.animation_delta_tolerance_info.pack(fill='both', expand=True,
                                                 padx=5, pady=5)
        self.animation_delta_tolerance_spinbox.pack(padx=5, pady=5, anchor="w")

        self.use_os_animation_count_limit_cbtn.pack(expand=True, fill='both')
	# TODO: Uncomment this once this works
        #self.calculate_limp_limb_vectors_cbtn.pack(expand=True, fill='both')


        self.apply_style()
        if self.app_root is not self:
            self.transient(self.app_root)

    def populate_animations_info_tree(self):
        jma_tree = self.jma_info_tree
        if not jma_tree['columns']:
            jma_tree['columns'] = ('data', )
            jma_tree.heading("#0")
            jma_tree.heading("data")
            jma_tree.column("#0", minwidth=100, width=100)
            jma_tree.column("data", minwidth=80, width=80, stretch=False)

        for iid in self._jma_tree_iids:
            jma_tree.delete(iid)

        self._jma_tree_iids = []

        if not self.jma_anims or not self.jma_anim_set:
            return

        nodes_iid = jma_tree.insert('', 'end', text="Nodes", tags=('item',),
                                    values=(len(self.jma_anim_set.nodes),))
        self._jma_tree_iids.append(nodes_iid)
        nodes = self.jma_anim_set.nodes
        for node in nodes:
            iid = jma_tree.insert(nodes_iid, 'end', text=node.name, tags=('item',))
            parent_name = child_name = sibling_name = "NONE"
            if node.sibling_index >= 0:
                sibling_name = nodes[node.sibling_index].name
            if node.first_child >= 0:
                child_name = nodes[node.first_child].name
            if node.parent_index >= 0:
                parent_name = nodes[node.parent_index].name

            jma_tree.insert(iid, 'end', text="Next sibling",
                            values=(sibling_name, ), tags=('item',),)
            jma_tree.insert(iid, 'end', text="First child",
                            values=(child_name, ), tags=('item',),)
            jma_tree.insert(iid, 'end', text="Parent",
                            values=(parent_name, ), tags=('item',),)


        anims_iid = jma_tree.insert('', 'end', text="Animations", tags=('item',),
                                    values=(len(self.jma_anims),))
        self._jma_tree_iids.append(anims_iid)
        for jma_anim in self.jma_anims:
            iid = jma_tree.insert(anims_iid, 'end', tags=('item',),
                                  text=jma_anim.name + jma_anim.ext)
            jma_tree.insert(iid, 'end', text="Node list checksum", tags=('item',),
                            values=(jma_anim.node_list_checksum, ))
            jma_tree.insert(iid, 'end', text="World relative", tags=('item',),
                            values=(jma_anim.world_relative, ))
            jma_tree.insert(iid, 'end', text="Type", tags=('item',),
                            values=(jma_anim.anim_type, ))
            jma_tree.insert(iid, 'end', text="Frame info", tags=('item',),
                            values=(jma_anim.frame_info_type, ))

            rot_flags   = jma_anim.rot_flags
            trans_flags = jma_anim.trans_flags
            scale_flags = jma_anim.scale_flags

            node_flags_iid = jma_tree.insert(
                iid, 'end', text="Transform flags", tags=('item',),
                values=(len(jma_anim.nodes),))
            for n in range(len(jma_anim.nodes)):
                node_iid = jma_tree.insert(
                    node_flags_iid, 'end', text=jma_anim.nodes[n].name,
                    tags=('item',), values=(
                        "*" if (rot_flags[n] or
                                trans_flags[n] or
                                scale_flags[n])
                        else "",))

                jma_tree.insert(node_iid, 'end', text="Rotation",
                                values=(rot_flags[n], ), tags=('item',))
                jma_tree.insert(node_iid, 'end', text="Position",
                                values=(trans_flags[n], ), tags=('item',))
                jma_tree.insert(node_iid, 'end', text="Scale",
                                values=(scale_flags[n], ), tags=('item',))

            # code below is very CPU and RAM intensive.
            # don't remove this continue unless debugging
            continue
            print("REMINDER TO REMOVE THIS DEBUG IN COMPILER WINDOW")

            has_dxdy = "dx" in jma_anim.frame_info_type
            has_dz   = "dz" in jma_anim.frame_info_type
            has_dyaw = "dyaw" in jma_anim.frame_info_type

            root_data_iid = jma_tree.insert(
                iid, 'end', text="Root node data", tags=('item',),
                values=(len(jma_anim.root_node_info),))
            for f in range(len(jma_anim.root_node_info)):
                if not has_dxdy and not has_dz and not has_dyaw:
                    break

                state = jma_anim.root_node_info[f]
                frame_iid = jma_tree.insert(
                    root_data_iid, 'end', tags=('item',),
                    text="frame%s" % f
                    )
                if has_dxdy:
                    jma_tree.insert(frame_iid, 'end', text="dx",
                                    values=(state.dx, ), tags=('item',),)
                    jma_tree.insert(frame_iid, 'end', text="dy",
                                    values=(state.dy, ), tags=('item',),)

                if has_dz:
                    jma_tree.insert(frame_iid, 'end', text="dz",
                                    values=(state.dz, ), tags=('item',),)

                if has_dyaw:
                    jma_tree.insert(frame_iid, 'end', text="dyaw",
                                    values=(state.dyaw, ), tags=('item',),)

                if has_dxdy:
                    jma_tree.insert(frame_iid, 'end', text="x",
                                    values=(state.x, ), tags=('item',),)
                    jma_tree.insert(frame_iid, 'end', text="y",
                                    values=(state.y, ), tags=('item',),)

                if has_dz:
                    jma_tree.insert(frame_iid, 'end', text="z",
                                    values=(state.z, ), tags=('item',),)

                if has_dyaw:
                    jma_tree.insert(frame_iid, 'end', text="yaw",
                                    values=(state.yaw, ), tags=('item',),)

            # even more CPU / RAM intensive code past here
            continue
            nodes_iid = jma_tree.insert(
                iid, 'end', text="Frame data", tags=('item',),
                values=(len(jma_anim.nodes),))
            for n in range(len(jma_anim.nodes)):
                states_iid = jma_tree.insert(
                    nodes_iid, 'end', text=jma_anim.nodes[n].name,
                    tags=('item',))

                for f in range(len(jma_anim.frames)):
                    state = jma_anim.frames[f][n]
                    node_iid = jma_tree.insert(
                        states_iid, 'end', tags=('item',),
                        text="frame%s" % f
                        )
                    jma_tree.insert(node_iid, 'end', text="i",
                                    values=(state.rot_i, ), tags=('item',),)
                    jma_tree.insert(node_iid, 'end', text="j",
                                    values=(state.rot_j, ), tags=('item',),)
                    jma_tree.insert(node_iid, 'end', text="k",
                                    values=(state.rot_k, ), tags=('item',),)
                    jma_tree.insert(node_iid, 'end', text="w",
                                    values=(state.rot_w, ), tags=('item',),)

                    jma_tree.insert(node_iid, 'end', text="x",
                                    values=(state.pos_x, ), tags=('item',),)
                    jma_tree.insert(node_iid, 'end', text="y",
                                    values=(state.pos_y, ), tags=('item',),)
                    jma_tree.insert(node_iid, 'end', text="z",
                                    values=(state.pos_z, ), tags=('item',),)

                    jma_tree.insert(node_iid, 'end', text="scale",
                                    values=(state.scale, ), tags=('item',),)


    def jma_dir_browse(self):
        if self._compiling or self._loading or self._saving:
            return

        tags_dir = self.tags_dir.get()
        # Add data to the path and then use path_replace to match the case of any
        # data directory that might already be here.
        data_dir = str(path_replace(Path(tags_dir).parent.joinpath("data"), "data", "data"))
        jma_dir = self.jma_dir.get()
        if tags_dir and not jma_dir:
            jma_dir = data_dir

        dirpath = path_normalize(askdirectory(
            initialdir=jma_dir, parent=self,
            title="Select the folder of animations to compile..."))

        if not dirpath:
            return

        dirpath = str(Path(dirpath))
        if tags_dir and data_dir and os.path.basename(dirpath).lower() == "animations":
            object_dir = os.path.dirname(dirpath)

            if object_dir and is_in_dir(object_dir, data_dir):
                tag_path = os.path.join(object_dir, os.path.basename(object_dir))
                tag_path = os.path.join(tags_dir, os.path.relpath(tag_path, data_dir))
                self.model_animations_path.set(tag_path + ".model_animations")

        self.app_root.last_load_dir = os.path.dirname(dirpath)
        self.jma_dir.set(dirpath)
        if not self.tags_dir.get():
            self.tags_dir.set(
                os.path.join(
                    path_split(self.app_root.last_load_dir, "data"),
                    "tags"))

    def tags_dir_browse(self):
        if self._compiling or self._loading or self._saving:
            return

        old_tags_dir = self.tags_dir.get()
        tags_dir = askdirectory(
            initialdir=old_tags_dir, parent=self,
            title="Select the root of the tags directory")

        if not tags_dir:
            return

        tags_dir = str(Path(tags_dir))

        antr_path = self.model_animations_path.get()
        if old_tags_dir and antr_path and not is_in_dir(antr_path, tags_dir):
            # adjust antr filepath to be relative to the new tags directory
            antr_path = os.path.join(tags_dir, os.path.relpath(antr_path, old_tags_dir))
            self.model_animations_path.set(antr_path)

        self.app_root.last_load_dir = os.path.dirname(tags_dir)
        self.tags_dir.set(tags_dir)

    def model_animations_path_browse(self, force=False):
        if not force and (self._compiling or self._loading or self._saving):
            return

        antr_dir = os.path.dirname(self.model_animations_path.get())
        if self.tags_dir.get() and not antr_dir:
            antr_dir = self.tags_dir.get()

        fp = asksaveasfilename(
            initialdir=antr_dir, title="Save model_animations to...", parent=self,
            filetypes=(("Model animations graph", "*.model_animations"), ('All', '*')))

        if not fp:
            return

        fp = Path(fp).with_suffix(".model_animations")

        self.app_root.last_load_dir = str(fp.parent)
        self.model_animations_path.set(str(fp))

        self.tags_dir.set(
            path_split(self.app_root.last_load_dir, "tags", after=True))

    def apply_style(self, seen=None):
        BinillaWidget.apply_style(self, seen)
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def set_animation_delta_tolerance(self):
        try:
            new_tolerance = float(self.animation_delta_tolerance_string.get())
            if new_tolerance >= 0:
                self.animation_delta_tolerance = new_tolerance
                return

            self.animation_delta_tolerance = 0
        except Exception:
            return

        self.animation_delta_tolerance_string.set(
            str(("%.20f" % self.animation_delta_tolerance)).rstrip("0").rstrip("."))

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        window_base_class.destroy(self)

    def load_animations(self):
        if not self._compiling and not self._loading and not self._saving:
            self._loading = True
            try:
                self._load_animations()
            except Exception:
                print(format_exc())
            try:
                self.populate_animations_info_tree()
            except Exception:
                print(format_exc())
            self._loading = False

    def save_animations(self):
        if not self._compiling and not self._loading and not self._saving:
            self._saving = True
            try:
                self._save_animations()
            except Exception:
                print(format_exc())
            self._saving = False

    def compile_model_animations(self):
        if not self._compiling and not self._loading and not self._saving:
            self._compiling = True
            try:
                self._compile_model_animations()
            except Exception:
                print(format_exc())
            self._compiling = False

    def _load_animations(self):
        animations_dir = self.jma_dir.get()
        if not animations_dir:
            return

        start = time.time()
        print("Locating jma files...")
        fps = []
        for _, __, files in os.walk(animations_dir):
            for fname in files:
                ext = os.path.splitext(fname)[-1].lower()
                if ext in JMA_ANIMATION_EXTENSIONS:
                    fps.append(os.path.join(animations_dir, fname))

            break

        if not fps:
            print("    No valid jma files found in the folder.")
            return

        self.jma_anim_set = None

        jma_anims = self.jma_anims = []
        print("Loading jma files...")
        self.update()
        for fp in fps:
            try:
                #print("    %s" % fp.replace('/', '\\').split("\\")[-1])
                self.update()

                anim_name = os.path.basename(fp)
                ext = os.path.splitext(fp)[-1].lower()

                jma_anim = None
                if ext in JMA_ANIMATION_EXTENSIONS:
                    with open(fp, "r") as f:
                        jma_anim = read_jma(f.read(), '', anim_name)

                if jma_anim:
                    jma_anims.append(jma_anim)
            except Exception:
                print(format_exc())
                print("    Could not parse '%s'" % anim_name)
                self.update()

        if not jma_anims:
            print("    No valid jma files found.")
            return

        first_crc = None
        for jma_anim in jma_anims:
            if first_crc is None:
                first_crc = jma_anim.node_list_checksum
            elif first_crc != jma_anim.node_list_checksum:
                print("    Warning, not all node list checksums match.")
                break


        print("Merging jma data...")
        self.app_root.update()
        self.jma_anim_set = JmaAnimationSet()
        errors_occurred = False
        for jma_anim in jma_anims:
            errors = self.jma_anim_set.merge_jma_animation(jma_anim)
            errors_occurred |= bool(errors)
            if errors:
                print("    Errors in '%s'" % jma_anim.name)
                for error in errors:
                    print("        ", error, sep='')

            self.update()

        antr_path = self.model_animations_path.get()
        if errors_occurred:
            print("    Errors occurred while loading jma files.")

        print("Finished loading animations. Took %s seconds.\n" %
              str(time.time() - start).split('.')[0])

    def _save_animations(self):
        animations_dir = self.jma_dir.get()
        if not animations_dir:
            return

        start = time.time()
        print("Saving jma animations...")
        self.update()
        for jma_anim in self.jma_anims:
            if isinstance(jma_anim, JmaAnimation):
                jma_filepath = os.path.join(
                    animations_dir, jma_anim.name + jma_anim.ext)
                write_jma(jma_filepath, jma_anim)

        print("Finished saving animations. Took %s seconds.\n" %
              str(time.time() - start).split('.')[0])

    def _compile_model_animations(self):
        if not self.jma_anim_set:
            return

        print("Compiling...")
        while not self.model_animations_path.get():
            self.model_animations_path_browse(True)
            if (not self.model_animations_path.get()) and self.warn_cancel():
                print("    Model_animations compilation cancelled.")
                return

        try:
            antr_tag = antr_def.build(filepath=self.model_animations_path.get())
        except Exception:
            antr_tag = None

        mod2_nodes = None
        if self.calculate_limp_limb_vectors.get():
            antr_path = self.model_animations_path.get()
            antr_dir = os.path.dirname(antr_path)
            mod2_path = os.path.join(
                antr_dir, os.path.splitext(os.path.basename(antr_path))[0] + ".gbxmodel")
            while mod2_nodes is None:
                try:
                    mod2_nodes = partial_mod2_def.build(filepath=mod2_path).\
                                 data.tagdata.nodes.STEPTREE
                    break
                except Exception:
                    print("Could not load the selected gbxmodel.")

                mod2_path = asksaveasfilename(
                    initialdir=antr_dir, parent=self,
                    title="Select the gbxmodel to get nodes from",
                    filetypes=(("Gearbox model", "*.gbxmodel"), ('All', '*')))

                if (not mod2_path) and self.warn_cancel():
                    print("    Model_animations compilation cancelled.")
                    return

        updating = antr_tag is not None
        if updating:
            print("Updating existing model_animations tag.")
        else:
            print("Creating new model_animations tag.")
            antr_tag = antr_def.build()

        antr_tag.filepath = self.model_animations_path.get()

        self.update()
        errors = compile_model_animations(antr_tag, self.jma_anim_set, False,
                                          self.animation_count_limit.get(),
                                          self.animation_delta_tolerance,
                                          self.update_mode.get(), mod2_nodes)
        if errors:
            for error in errors:
                print(error)

            self.update()
            if not messagebox.askyesno(
                    "Model_animations compilation failed",
                    "Errors occurred while compiling animations(check console). "
                    "Do you want to save the model_animations tag anyway?",
                    icon='warning', parent=self):
                print("    Model_animations compilation failed.")
                return

        try:
            antr_tag.calc_internal_data()
            antr_tag.serialize(temp=False, backup=False, calc_pointers=False,
                               int_test=False)
            print("    Finished")
        except Exception:
            print(format_exc())
            print("    Could not save compiled model_animations.")

    def warn_cancel(self):
        return bool(messagebox.askyesno(
            "Unsaved model_animations",
            "Are you sure you wish to cancel?",
            icon='warning', parent=self))


if __name__ == "__main__":
    AnimationsCompilerWindow(None).mainloop()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

# ##############
#   metadata   #
# ##############
__author__ = "Devin Bobadilla, Michelle van der Graaf"
#           YYYY.MM.DD
__date__ = "2020.02.29"
__version__ = (1, 9, 0)
__website__ = "https://github.com/Sigmmma/mozzarilla"



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase

from pathlib import PurePath
import threadsafe_tkinter as tk

from math import sqrt
from traceback import format_exc

from reclaimer.hek.defs.mod2    import fast_mod2_def as mod2_def
from reclaimer.stubbs.defs.mode import fast_mode_def as mode_def

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


def convert_model(src_tag, dst_tag, to_gbxmodel):
    src_tag_data = src_tag.data.tagdata
    dst_tag_data = dst_tag.data.tagdata

    # move the first 14 header fields from src tag into dst tag
    # (except for the flags since usually ZONER shouldnt be copied)
    dst_tag_data[1: 14] = src_tag_data[1: 14]
    for flag_name in src_tag_data.flags.NAME_MAP:
        if hasattr(dst_tag_data.flags, flag_name):
            dst_tag_data.flags[flag_name] = src_tag_data.flags[flag_name]

    # fix the fact the mode and mod2 store stuff related to lods
    # in reverse on most platforms(pc stubbs is an exception)
    if dst_tag_data.superhigh_lod_cutoff < dst_tag_data.superlow_lod_cutoff:
        tmp0 = dst_tag_data.superhigh_lod_cutoff
        tmp1 = dst_tag_data.high_lod_cutoff
        dst_tag_data.superhigh_lod_cutoff = dst_tag_data.superlow_lod_cutoff
        dst_tag_data.high_lod_cutoff      = dst_tag_data.low_lod_cutoff
        dst_tag_data.low_lod_cutoff       = tmp1
        dst_tag_data.superlow_lod_cutoff  = tmp0

        tmp0 = dst_tag_data.superhigh_lod_nodes
        tmp1 = dst_tag_data.high_lod_nodes
        dst_tag_data.superhigh_lod_nodes = dst_tag_data.superlow_lod_nodes
        dst_tag_data.high_lod_nodes      = dst_tag_data.low_lod_nodes
        dst_tag_data.low_lod_nodes       = tmp1
        dst_tag_data.superlow_lod_nodes  = tmp0

    # make all markers global ones
    if hasattr(src_tag, "globalize_local_markers"):
        src_tag.globalize_local_markers()

    # move the markers, nodes, regions, and shaders, from mode into mod2
    dst_tag_data.markers = src_tag_data.markers
    dst_tag_data.nodes = src_tag_data.nodes
    dst_tag_data.regions = src_tag_data.regions
    dst_tag_data.shaders = src_tag_data.shaders

    # give the mod2 as many geometries as the mode
    src_tag_geoms = src_tag_data.geometries.STEPTREE
    dst_tag_geoms = dst_tag_data.geometries.STEPTREE
    dst_tag_geoms.extend(len(src_tag_geoms))

    # copy the data from the src_tag_geoms into the dst_tag_geoms
    for i in range(len(dst_tag_geoms)):
        # give the dst_tag_geom as many parts as the src_tag_geom
        src_tag_parts = src_tag_geoms[i].parts.STEPTREE
        dst_tag_parts = dst_tag_geoms[i].parts.STEPTREE
        dst_tag_parts.extend(len(src_tag_parts))

        # copy the data from the src_tag_parts into the dst_tag_parts
        for j in range(len(dst_tag_parts)):
            src_tag_part = src_tag_parts[j]
            dst_tag_part = dst_tag_parts[j]

            # move the first 9 part fields from src_tag into dst_tag
            # (except for the flags since usually ZONER shouldnt be copied)
            dst_tag_part[1: 9] = src_tag_part[1: 9]

            src_local_nodes = getattr(src_tag_part, "local_nodes", None)
            dst_local_nodes = getattr(dst_tag_part, "local_nodes", None)
            if not getattr(src_tag_part.flags, "ZONER", False):
                src_local_nodes = None

            if dst_local_nodes and src_local_nodes:
                # converting from a gbxmodel with local nodes to a gbxmodel
                # with local nodes. copy the local nodes and node count
                dst_tag_part.flags.ZONER = True
                dst_tag_part.local_node_count = src_tag_part.local_node_count
                dst_tag_part.local_nodes[:] = src_local_nodes[:]
            elif src_local_nodes:
                # converting from a gbxmodel with local nodes to
                # something without them. make the nodes absolute.
                src_tag.delocalize_part_nodes(i, j)

            # move the vertices and triangles from the src_tag into the dst_tag
            dst_tag_part.triangles = src_tag_part.triangles
            dst_tag_part.uncompressed_vertices = src_tag_part.uncompressed_vertices
            dst_tag_part.compressed_vertices   = src_tag_part.compressed_vertices

            uncomp_verts = dst_tag_part.uncompressed_vertices
            comp_verts   = dst_tag_part.compressed_vertices

            if to_gbxmodel:
                # if the compressed vertices are valid or
                # the uncompressed are not then we don't have
                # any conversion to do(already uncompressed)
                if not uncomp_verts.size or comp_verts.size:
                    dst_tag.decompress_part_verts(i, j)
            elif not comp_verts.size or uncomp_verts.size:
                # the uncompressed vertices are valid or
                # the compressed are not, so we don't have
                # any conversion to do(already compressed)
                dst_tag.compress_part_verts(i, j)

    dst_tag.calc_internal_data()


class ModelConverter(ConverterBase, window_base_class):
    to_gbxmodel = True
    src_ext = "model"
    dst_ext = "gbxmodel"

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        self.setup_window(*args, **kwargs)

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)
        self.pack_widgets()
        self.apply_style()

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def convert(self, tag_path):
        if self.to_gbxmodel:
            src_tag = mode_def.build(filepath=tag_path)
            dst_tag = mod2_def.build()
        else:
            src_tag = mod2_def.build(filepath=tag_path)
            dst_tag = mode_def.build()

        dst_tag.filepath = PurePath(tag_path).with_suffix("." + self.dst_ext)
        convert_model(src_tag, dst_tag, self.to_gbxmodel)
        return dst_tag


if __name__ == "__main__":
    try:
        ModelConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from traceback import format_exc
try:
    from . import model_converter
except (ImportError, SystemError):
    import model_converter


model_converter.window_base_class = model_converter.tk.Toplevel
if __name__ == "__main__":
    model_converter.window_base_class = model_converter.tk.Tk


class GbxmodelConverter(model_converter.ConverterBase,
                        model_converter.window_base_class):
    to_gbxmodel = False
    src_ext = "gbxmodel"
    dst_ext = "model"

    __init__ = model_converter.ModelConverter.__init__
    setup_window = model_converter.ModelConverter.setup_window
    apply_style = model_converter.ModelConverter.apply_style
    destroy = model_converter.ModelConverter.destroy
    convert = model_converter.ModelConverter.convert


if __name__ == "__main__":
    try:
        GbxmodelConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

from pathlib import Path, PureWindowsPath
import os
import tkinter as tk
import zipfile

from threading import Thread
from traceback import format_exc

from binilla.widgets.binilla_widget import BinillaWidget
from binilla.windows.filedialog import askopenfilename, asksaveasfilename

from supyr_struct.util import path_normalize, is_in_dir, tagpath_to_fullpath
from mozzarilla.widgets.directory_frame import DirectoryFrame,\
     HierarchyFrame, DependencyFrame
from mozzarilla import editor_constants as e_c


class DependencyWindow(tk.Toplevel, BinillaWidget):
    app_root = None
    handler = None

    _zipping = False
    stop_zipping = False

    def __init__(self, app_root, *args, **kwargs):
        self.handler = app_root.handler
        self.app_root = app_root
        kwargs.update(width=400, height=500, bd=0,
                      highlightthickness=0, bg=self.default_bg_color)
        tk.Toplevel.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)
        self.title("[%s] Tag dependency viewer / zipper" %
                   app_root.handler_names[app_root._curr_handler_index])
        self.minsize(width=400, height=100)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        # make the tkinter variables
        self.tag_filepath = tk.StringVar(self)

        # make the frames
        self.filepath_frame = tk.LabelFrame(self, text="Select a tag")
        self.button_frame = tk.LabelFrame(self, text="Actions")

        self.display_button = tk.Button(
            self.button_frame, width=25, text='Show dependencies',
            command=self.populate_dependency_tree)

        self.zip_button = tk.Button(
            self.button_frame, width=25, text='Zip tag recursively',
            command=self.recursive_zip)

        self.dependency_frame = DependencyFrame(self, app_root=self.app_root)

        self.filepath_entry = tk.Entry(
            self.filepath_frame, textvariable=self.tag_filepath)
        self.browse_button = tk.Button(
            self.filepath_frame, text="Browse", command=self.browse)

        self.display_button.pack(padx=4, pady=2, side='left')
        self.zip_button.pack(padx=4, pady=2, side='right')

        self.filepath_entry.pack(padx=(4, 0), pady=2, side='left',
                                 expand=True, fill='x')
        self.browse_button.pack(padx=(0, 4), pady=2, side='left')

        self.filepath_frame.pack(fill='x', padx=1)
        self.button_frame.pack(fill='x', padx=1)
        self.dependency_frame.pack(fill='both', padx=1, expand=True)

        self.transient(app_root)
        self.apply_style()
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def browse(self):
        if self._zipping:
            return

        filetypes = [('All', '*')]

        defs = self.app_root.handler.defs
        for def_id in sorted(defs.keys()):
            filetypes.append((def_id, defs[def_id].ext))
        fp = askopenfilename(
            title="Select a tag", filetypes=filetypes, parent=self,
            initialdir=self.app_root.last_load_dir)

        if not fp:
            return

        self.app_root.last_load_dir = Path(fp).parent
        self.tag_filepath.set(fp)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        self.stop_zipping = True
        tk.Toplevel.destroy(self)

    def get_tag(self, tag_path):
        try:
            return self.handler.get_tag(tag_path, load_unloaded=True)
        except KeyError:
            print(format_exc())
            return None

    def get_dependencies(self, tag):
        def_id = tag.def_id
        dependency_cache = self.handler.tag_ref_cache.get(def_id)

        if not dependency_cache:
            return ()

        nodes = self.handler.get_nodes_by_paths(
            self.handler.tag_ref_cache[def_id], tag.data)

        dependencies = []

        for node in nodes:
            # if the node's filepath is empty, just skip it
            if not node.filepath:
                continue

            ext = '.' + node.tag_class.enum_name

            if tagpath_to_fullpath(
                self.handler.tagsdir, PureWindowsPath(node.filepath), extension=ext
            ) is not None and (self.handler.treat_mode_as_mod2 and ext == '.model'):
                ext = '.gbxmodel'

            dependencies.append(node.filepath + ext)
        return dependencies

    def populate_dependency_tree(self):
        filepath = self.tag_filepath.get()
        if not filepath:
            return

        app = self.app_root
        handler = self.handler = app.handler
        handler_name = app.handler_names[app._curr_handler_index]
        if handler_name not in app.tags_dir_relative:
            print("Change the current tag set.")
            return

        filepath = path_normalize(filepath)

        if not is_in_dir(filepath, self.handler.tagsdir):
            print("%s\nis not in tagsdir\n%s" %
                  (filepath, self.handler.tagsdir))
            return

        rel_filepath = Path(filepath).relative_to(self.handler.tagsdir)
        tag = self.get_tag(rel_filepath)
        if tag is None:
            print("Could not load tag:\n    %s" % filepath)
            return

        self.dependency_frame.handler = handler
        self.dependency_frame.tags_dir = self.handler.tagsdir
        self.dependency_frame.root_tag_path = tag.filepath
        self.dependency_frame.root_tag_text = rel_filepath

        self.dependency_frame.reload()

    def recursive_zip(self):
        if self._zipping:
            return
        try: self.zip_thread.join()
        except Exception: pass
        self.zip_thread = Thread(target=self._recursive_zip)
        self.zip_thread.daemon = True
        self.zip_thread.start()

    def _recursive_zip(self):
        self._zipping = True
        try:
            self.do_recursive_zip()
        except Exception:
            print(format_exc())
        self._zipping = False

    def do_recursive_zip(self):
        tag_path = self.tag_filepath.get()
        if not tag_path:
            return

        app = self.app_root
        handler = self.handler = app.handler
        handler_name = app.handler_names[app._curr_handler_index]
        if handler_name not in app.tags_dir_relative:
            print("Change the current tag set.")
            return

        tag_path = Path(tag_path)
        if not is_in_dir(tag_path, self.handler.tagsdir):
            print("Specified tag is not located within the tags directory")
            return

        tagzip_path = asksaveasfilename(
            initialdir=self.app_root.last_load_dir, parent=self,
            title="Save zipfile to...", filetypes=(("zipfile", "*.zip"), ))

        if not tagzip_path:
            return

        try:
            rel_filepath = tag_path.relative_to(self.handler.tagsdir)
            tag = self.get_tag(rel_filepath)
        except ValueError:
            tag = None

        if tag is None:
            print("Could not load tag:\n    %s" % tag_path)
            return

        # make the zipfile to put everything in
        tagzip_path = os.path.splitext(tagzip_path)[0] + ".zip"

        tags_to_zip = [rel_filepath]
        new_tags_to_zip = []
        seen_tags = set()

        with zipfile.ZipFile(str(tagzip_path), mode='w') as tagzip:
            # loop over all the tags and add them to the zipfile
            while tags_to_zip:
                for rel_filepath in tags_to_zip:
                    tag_path = tagpath_to_fullpath(
                        self.handler.tagsdir, PureWindowsPath(rel_filepath))
                    if self.stop_zipping:
                        print('Recursive zip operation cancelled.\n')
                        return

                    if rel_filepath in seen_tags:
                        continue
                    seen_tags.add(rel_filepath)

                    try:
                        print("Adding '%s' to zipfile" % rel_filepath)
                        app.update_idletasks()
                        tag = self.get_tag(rel_filepath)
                        new_tags_to_zip.extend(self.get_dependencies(tag))

                        # try to conserve memory a bit
                        del tag

                        tagzip.write(str(tag_path), arcname=str(rel_filepath))
                    except Exception:
                        print(format_exc())
                        print("    Could not add '%s' to zipfile." %
                              rel_filepath)

                # replace the tags to zip with the newly collected ones
                tags_to_zip[:] = new_tags_to_zip
                del new_tags_to_zip[:]

        print("\nRecursive zip completed.\n")



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import ctypes
import gc
import os
import sys
import tkinter as tk
import weakref

from copy import deepcopy
from pathlib import Path
from threading import Thread
from time import time
from traceback import format_exc

import arbytmap as ab

from reclaimer.bitmaps.p8_palette import HALO_P8_PALETTE, STUBBS_P8_PALETTE
from reclaimer.hek.defs.bitm import bitm_def
from reclaimer.constants import TYPE_NAME_MAP, FORMAT_NAME_MAP,\
     I_FORMAT_NAME_MAP

from binilla.util import do_subprocess, ProcController
from binilla.widgets.binilla_widget import BinillaWidget
from binilla.widgets.scroll_menu import ScrollMenu
from binilla.windows.filedialog import askdirectory, asksaveasfilename
from mozzarilla.widgets.field_widgets import HaloBitmapDisplayFrame,\
     HaloBitmapDisplayBase
from mozzarilla import editor_constants as e_c

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk


#                      (A, R, G, B)
PC_ARGB_TO_XBOX_ARGB = (1, 3, 2, 0)
XBOX_ARGB_TO_PC_ARGB = (3, 0, 2, 1)

AL_COMBO_TO_AL   = (0, 0)
AL_COMBO_TO_ARGB = (0, 0, 0, 0)


BITMAP_PLATFORMS = ("PC", "XBOX")
MULTI_SWAP_OPTIONS = ("", "PC to XBOX", "XBOX to PC")
AY8_OPTIONS = ("Alpha", "Intensity")
EXTRACT_TO_OPTIONS = ("", "DDS", "TGA", "PNG")
NO_YES_OPTIONS = ("No", "Yes")
BITMAP_TYPES = TYPE_NAME_MAP
BITMAP_FORMATS = FORMAT_NAME_MAP
DXT_ALPHA_FORMATS = (ab.FORMAT_DXT2, ab.FORMAT_DXT3,
                     ab.FORMAT_DXT4, ab.FORMAT_DXT5)

VALID_FORMAT_ENUMS = (0, 1, 2, 3, 6, 8, 9, 10, 11, 14, 15, 16, 17)
PARAM_FORMAT_TO_FORMAT = (-1, ) + VALID_FORMAT_ENUMS
FORMAT_OPTIONS = ("Unchanged", ) + tuple(BITMAP_FORMATS[i] for i in VALID_FORMAT_ENUMS)

HALO_1_TYPE_COUNT   = 4
HALO_1_FORMAT_COUNT = 18


platform = sys.platform.lower()
if "linux" in platform:
    platform = "linux"


SetFileAttributesW = None
if platform == "win32":
    TEXT_EDITOR_NAME = "notepad"
    SetFileAttributesW = ctypes.windll.kernel32.SetFileAttributesW
elif platform == "darwin":
    TEXT_EDITOR_NAME = "TextEdit"
elif platform == "linux":
    TEXT_EDITOR_NAME = "vim"
else:
    # idfk
    TEXT_EDITOR_NAME = "nano"


class ConversionFlags:
    platform = BITMAP_PLATFORMS.index("PC")
    multi_swap = MULTI_SWAP_OPTIONS.index("")
    mono_channel_to_keep = AY8_OPTIONS.index("Alpha")

    extract_to = EXTRACT_TO_OPTIONS.index("")
    downres = 0
    alpha_bias = 127
    new_format = 0

    prune_tiff = 0
    swizzled = 0
    mono_swap = 0
    ck_trans = 0
    mip_gen = 0

    extract_path = ""


class BitmapInfo:
    type = 0
    format = 0
    swizzled = False
    width = 0
    height = 0
    depth = 0
    mipmaps = 0

    def __init__(self, bitmap_block=None):
        if not bitmap_block:
            return
        self.type = bitmap_block.type.data
        self.format = bitmap_block.format.data
        self.swizzled = bool(bitmap_block.flags.swizzled)
        self.width = bitmap_block.width
        self.height = bitmap_block.height
        self.depth = bitmap_block.depth
        self.mipmaps = bitmap_block.mipmaps


class BitmapTagInfo:
    platform = 0
    pixel_data_size = 0
    tiff_data_size = 0
    bitmap_infos = ()

    def __init__(self, bitm_tag=None):
        self.bitmap_infos = []
        if not bitm_tag:
            return

        bitm_data = bitm_tag.data.tagdata
        self.tiff_data_size = bitm_data.compressed_color_plate_data.size
        self.pixel_data_size = bitm_data.processed_pixel_data.size
        for bitmap in bitm_data.bitmaps.STEPTREE:
            self.bitmap_infos.append(BitmapInfo(bitmap))

        self.platform = bitm_tag.is_xbox_bitmap

    @property
    def type(self):
        return 0 if not self.bitmap_infos else self.bitmap_infos[0].type
    @property
    def format(self):
        return 0 if not self.bitmap_infos else self.bitmap_infos[0].format
    @property
    def swizzled(self):
        return 0 if not self.bitmap_infos else self.bitmap_infos[0].swizzled


def get_will_be_converted(flags, tag_info):
    if flags.platform != tag_info.platform:
        return True
    elif (flags.swizzled != tag_info.swizzled and
          tag_info.format not in (14, 15, 16)):
        return True
    elif (flags.mono_swap and (PARAM_FORMAT_TO_FORMAT[flags.new_format] == 3 or
                               tag_info.format == 3)):
        return True
    elif flags.multi_swap and tag_info.format in (6, 8, 9, 10, 11, 14, 15, 16):
        return True

    for info in tag_info.bitmap_infos:
        if PARAM_FORMAT_TO_FORMAT[flags.new_format] not in (-1, info.format):
            return True
        elif flags.downres and max(info.width, info.height, info.depth) > 4:
            return True
        elif flags.mip_gen and info.mipmaps == 0:
            return True

    return False


def get_channel_mappings(conv_flags, bitmap_info):
    mono_swap = conv_flags.mono_swap
    fmt_s = FORMAT_NAME_MAP[bitmap_info.format]
    fmt_t = PARAM_FORMAT_TO_FORMAT[conv_flags.new_format]
    fmt_t = fmt_s if fmt_t < 0 else FORMAT_NAME_MAP[fmt_t]

    multi_swap = conv_flags.multi_swap
    chan_to_keep = conv_flags.mono_channel_to_keep

    chan_ct_s = ab.CHANNEL_COUNTS[fmt_s]
    chan_ct_t = ab.CHANNEL_COUNTS[fmt_t]
    chan_map = None
    chan_merge_map = None

    if chan_ct_s == 4:
        if chan_ct_t == 4:
            # TAKES CARE OF ALL THE MULTIPURPOSE CHANNEL SWAPPING
            if multi_swap == 1:
                chan_map = PC_ARGB_TO_XBOX_ARGB
            elif multi_swap == 2:
                chan_map = XBOX_ARGB_TO_PC_ARGB

        elif fmt_t == ab.FORMAT_A8L8:
            chan_merge_map = ab.M_ARGB_TO_LA if mono_swap else ab.M_ARGB_TO_AL

        elif fmt_t in (ab.FORMAT_A8, ab.FORMAT_L8, ab.FORMAT_AL8):
            # CONVERTING FROM A 4 CHANNEL FORMAT TO MONOCHROME
            if fmt_t == ab.FORMAT_L8:
                chan_merge_map = ab.M_ARGB_TO_L
            elif fmt_t == ab.FORMAT_A8 or chan_to_keep == 0:
                chan_map = ab.ANYTHING_TO_A
                chan_merge_map = ab.M_ARGB_TO_A
            else:
                chan_merge_map = ab.M_ARGB_TO_L

    elif chan_ct_s == 2:
        # CONVERTING FROM A 2 CHANNEL FORMAT TO OTHER FORMATS
        if fmt_s == ab.FORMAT_AL8:
            chan_map = AL_COMBO_TO_ARGB if chan_ct_t == 4 else AL_COMBO_TO_AL

        elif fmt_s == ab.FORMAT_A8L8:
            if chan_ct_t == 4:
                if mono_swap:
                    chan_map = ab.LA_TO_ARGB
                else:
                    chan_map = ab.AL_TO_ARGB
            elif fmt_t == ab.FORMAT_A8 or (fmt_t == ab.FORMAT_AL8 and
                                           not chan_to_keep):
                chan_map = ab.ANYTHING_TO_A
            elif fmt_t == ab.FORMAT_L8 or (fmt_t == ab.FORMAT_AL8 and
                                           chan_to_keep):
                chan_map = ab.AL_TO_L
            elif mono_swap:
                if fmt_t == ab.FORMAT_A8L8:
                    chan_map = ab.AL_TO_LA
                elif chan_ct_t == 4:
                    chan_map = ab.LA_TO_ARGB

    elif chan_ct_s == 1:
        # CONVERTING FROM A 1 CHANNEL FORMAT TO OTHER FORMATS
        if chan_ct_t == 4:
            if fmt_s == ab.FORMAT_A8:
                chan_map = ab.A_TO_ARGB
            elif fmt_s == ab.FORMAT_L8:
                chan_map = ab.L_TO_ARGB

        elif chan_ct_t == 2:
            if fmt_s == ab.FORMAT_A8:
                chan_map = ab.A_TO_AL
            elif fmt_s == ab.FORMAT_L8:
                chan_map = ab.L_TO_AL

    return chan_map, chan_merge_map


def convert_bitmap_tag(tag, conv_flags, bitmap_info, use_stubbs_p8=False):
    for i in range(tag.bitmap_count()):
        if not tag.is_power_of_2_bitmap(i):
            return False

    new_format = BITMAP_FORMATS[
        PARAM_FORMAT_TO_FORMAT[conv_flags.new_format]]

    extract_ext = EXTRACT_TO_OPTIONS[conv_flags.extract_to]
    ck_trans = conv_flags.ck_trans

    do_conversion = get_will_be_converted(conv_flags, bitmap_info)
    if not do_conversion and not extract_ext:
        return True

    arb = ab.Arbytmap()
    if tag.sanitize_mipmap_counts():
        print("ERROR: Bad mipmap counts in:\n%s\t\n" % tag.filepath)
        return False

    tag.parse_bitmap_blocks()
    pixel_data = tag.data.tagdata.processed_pixel_data.data

    for i in range(tag.bitmap_count()):
        typ   = BITMAP_TYPES[tag.bitmap_type(i)]
        fmt_s = BITMAP_FORMATS[tag.bitmap_format(i)]
        fmt_t = fmt_s if conv_flags.new_format <= 0 else new_format

        #get the texture block to be loaded
        tex_block = list(pixel_data[i])
        tex_info = tag.tex_infos[i]

        if fmt_t == ab.FORMAT_P8_BUMP and typ in (ab.TYPE_CUBEMAP, ab.TYPE_3D):
            print("Cannot convert cubemaps or 3d textures to P8.")
            fmt_t = fmt_s
        elif fmt_t in ab.DDS_FORMATS and typ == ab.TYPE_3D:
            print("Cannot convert 3D textures to DXT formats.")
            fmt_t = fmt_s

        if (fmt_s in (ab.FORMAT_A8, ab.FORMAT_L8, ab.FORMAT_AL8) and
            fmt_t in (ab.FORMAT_A8, ab.FORMAT_L8, ab.FORMAT_AL8)):
            tex_info["format"] = fmt_s = fmt_t

        chan_map, chan_merge_map = get_channel_mappings(conv_flags, bitmap_info)
        palette_picker = None
        palettize = (fmt_t == ab.FORMAT_P8_BUMP)

        p8_palette = STUBBS_P8_PALETTE if use_stubbs_p8 else HALO_P8_PALETTE

        if "palette" in tex_info:
            tex_info["palette"] = [
                p8_palette.p8_palette_32bit_packed] * len(tex_block)

        # we want to preserve the color key transparency of
        # the original image if converting to the same format
        if fmt_s == fmt_t and fmt_t in (ab.FORMAT_P8_BUMP, ab.FORMAT_DXT1):
            # also need to make sure channels aren't being swapped around
            if not conv_flags.multi_swap:
                ck_trans = True

        if ab.CHANNEL_COUNTS[fmt_s] == 4:
            if not ck_trans or fmt_s in (ab.FORMAT_X8R8G8B8, ab.FORMAT_R5G6B5):
                palette_picker = p8_palette.argb_array_to_p8_array_best_fit
            else:
                palette_picker = p8_palette.argb_array_to_p8_array_best_fit_alpha

        arb.load_new_texture(texture_block=tex_block, texture_info=tex_info)

        # build the initial conversion settings list from the above settings
        conv_settings = dict(
            swizzle_mode=conv_flags.swizzled, palettize=palettize,
            one_bit_bias=conv_flags.alpha_bias, palette_picker=palette_picker,
            downres_amount=conv_flags.downres, target_format=fmt_t,
            color_key_transparency=ck_trans, mipmap_gen=conv_flags.mip_gen,
            channel_mapping=chan_map, channel_merge_mapping=chan_merge_map)

        arb.load_new_conversion_settings(**conv_settings)

        if extract_ext and conv_flags.extract_path:
            path = conv_flags.extract_path
            if tag.bitmap_count() > 1:
                path = os.path.join(path, str(i))
            arb.save_to_file(output_path=path, ext=extract_ext)

        if do_conversion:
            success = arb.convert_texture()
            tag.tex_infos[i] = arb.texture_info  # tex_info may have changed

            if success:
                tex_root = pixel_data[i]
                tex_root.parse(initdata=arb.texture_block,
                               clear=False, init_attrs=False)
                tag.swizzled(i, arb.swizzled)

                #change the bitmap format to the new format
                tag.bitmap_format(i, I_FORMAT_NAME_MAP[arb.format])
            else:
                print("Error occurred while converting:\n\t%s\n" % tag.filepath)
                return False

    if do_conversion:
        tag.sanitize_bitmaps()
        tag.set_platform(conv_flags.platform)
        tag.add_bitmap_padding(conv_flags.platform)

    return True


class BitmapConverterWindow(window_base_class, BinillaWidget):
    app_root = None
    tag_list_frame = None
    loaded_tags_dir = ''
    last_load_dir = ''

    use_stubbs_p8 = None
    read_only = None
    backup_tags = None
    open_log = None

    conversion_flags = ()
    bitmap_tag_infos = ()
    bitmap_display_windows = ()

    _processing = False
    _cancel_processing = False
    _populating_settings = False
    _populating_bitmap_info = False
    _settings_enabled = True

    print_interval = 5

    # these cache references to the settings widgets for iteratively
    # enabling/disabling settings before and after converting.
    checkbuttons = ()
    buttons = ()
    spinboxes = ()
    menus = ()

    bitm_def = bitm_def

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)
            self.app_root = app_root
        else:
            self.app_root = self

        window_base_class.__init__(self, app_root, *args, **kwargs)
        BinillaWidget.__init__(self, app_root, *args, **kwargs)

        self.conversion_flags = {}
        self.bitmap_tag_infos = {}
        self.bitmap_display_windows = {}

        self.title("Bitmap converter")
        self.resizable(0, 1)
        self.update()
        try:
            self.iconbitmap(e_c.MOZZ_ICON_PATH)
        except Exception:
            print("Could not load window icon.")

        # make the tkinter variables
        self.read_only = tk.BooleanVar(self)
        self.backup_tags = tk.BooleanVar(self, True)
        self.open_log = tk.BooleanVar(self, True)
        self.use_stubbs_p8 = tk.BooleanVar(self)

        self.scan_dir_path = tk.StringVar(self)
        self.data_dir_path = tk.StringVar(self)
        self.log_file_path = tk.StringVar(self)

        self.write_trace(self.read_only, lambda *a, s=self:
                         s.convert_button.config(
                             text="Make log" if s.read_only.get()
                             else "Convert"))

        # make the frames
        self.main_frame = tk.Frame(self)
        self.settings_frame = tk.LabelFrame(self.main_frame, text="Settings")
        self.bitmap_info_frame = tk.LabelFrame(self.main_frame, text="Bitmap info")
        self.buttons_frame = tk.Frame(self.main_frame)
        self.tagset_info_frame = tk.Frame(self.main_frame)
        self.tag_list_frame = BitmapConverterList(self)

        self.log_file_frame  = tk.LabelFrame(
            self.settings_frame, text="Scan log filepath")
        self.scan_dir_frame = tk.LabelFrame(
            self.settings_frame, text="Directory to scan")
        self.data_dir_frame = tk.LabelFrame(
            self.settings_frame, text="Directory to extract data to")
        self.global_params_frame = tk.LabelFrame(
            self.settings_frame, text="Global parameters")
        self.params_frame = tk.Frame(self.settings_frame)


        self.general_params_frame = tk.LabelFrame(
            self.params_frame, text="General parameters")
        self.format_params_frame = tk.LabelFrame(
            self.params_frame, text="Format parameters")

        self.bitmap_index_frame = tk.Frame(self.bitmap_info_frame)


        self.curr_bitmap_label_0 = tk.Label(self.bitmap_index_frame, text="Current")
        self.curr_bitmap_spinbox = tk.Spinbox(self.bitmap_index_frame, width=3,
                                              state="readonly", repeatinterval=5,
                                              command=lambda *a, s=self:
                                              s.populate_bitmap_info())
        self.curr_bitmap_label_1 = tk.Label(self.bitmap_index_frame, text=" to ")
        self.max_bitmap_entry = tk.Entry(self.bitmap_index_frame, width=3,
                                         state='disabled')

        self.curr_type_menu = ScrollMenu(self.bitmap_info_frame, menu_width=8,
                                         disabled=True, options=BITMAP_TYPES)
        self.curr_format_menu = ScrollMenu(self.bitmap_info_frame, menu_width=8,
                                           disabled=True, options=BITMAP_FORMATS)
        self.curr_swizzled_menu = ScrollMenu(self.bitmap_info_frame, menu_width=8,
                                             options=NO_YES_OPTIONS, disabled=True)
        self.curr_platform_menu = ScrollMenu(self.bitmap_info_frame, menu_width=8,
                                             disabled=True, options=BITMAP_PLATFORMS)
        self.curr_has_tiff_menu = ScrollMenu(self.bitmap_info_frame, menu_width=8,
                                             options=NO_YES_OPTIONS, disabled=True)
        self.curr_height_entry = tk.Entry(self.bitmap_info_frame, width=12,
                                          state='disabled')
        self.curr_width_entry = tk.Entry(self.bitmap_info_frame, width=12,
                                         state='disabled')
        self.curr_depth_entry = tk.Entry(self.bitmap_info_frame, width=12,
                                         state='disabled')
        self.curr_mip_entry = tk.Entry(self.bitmap_info_frame, width=12,
                                       state='disabled')

        self.log_file_entry = tk.Entry(
            self.log_file_frame, textvariable=self.log_file_path,
            state=tk.DISABLED)
        self.log_file_browse_button = tk.Button(
            self.log_file_frame, text="Browse", command=self.log_browse)

        self.scan_dir_entry = tk.Entry(
            self.scan_dir_frame, textvariable=self.scan_dir_path,
            state=tk.DISABLED)
        self.scan_dir_browse_button = tk.Button(
            self.scan_dir_frame, text="Browse", command=self.scan_dir_browse)

        self.data_dir_entry = tk.Entry(
            self.data_dir_frame, textvariable=self.data_dir_path,
            state=tk.DISABLED)
        self.data_dir_browse_button = tk.Button(
            self.data_dir_frame, text="Browse", command=self.data_dir_browse)


        self.read_only_cbutton = tk.Checkbutton(
            self.global_params_frame, text="Read-only �",
            variable=self.read_only, command=self.update_all_path_colors)
        self.backup_tags_cbutton = tk.Checkbutton(
            self.global_params_frame, text="Backup tags �",
            variable=self.backup_tags)
        self.open_log_cbutton = tk.Checkbutton(
            self.global_params_frame, text="Show log �",
            variable=self.open_log)
        self.use_stubbs_p8_cbutton = tk.Checkbutton(
            self.global_params_frame, text="Use Stubbs p8 palette �",
            variable=self.use_stubbs_p8)


        self.read_only_cbutton.tooltip_string = (
            "Does no conversion, and instead writes a\n"
            "log detailing all bitmaps in the directory.")
        self.backup_tags_cbutton.tooltip_string = (
            "Backs up all bitmaps before editing\n"
            "(does nothing if a backup already exists)")
        self.open_log_cbutton.tooltip_string = (
            "Open the conversion log when finished.")
        self.use_stubbs_p8_cbutton.tooltip_string = (
            "Use Stubbs the Zombie's p8-bump palette\n"
            "instead of Halo's for P8-bump textures.")


        self.platform_menu = ScrollMenu(
            self.general_params_frame, menu_width=12,
            options=BITMAP_PLATFORMS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.platform_menu, "platform"))
        self.format_menu = ScrollMenu(
            self.general_params_frame, menu_width=12,
            options=FORMAT_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.format_menu, "new_format"))
        self.extract_to_menu = ScrollMenu(
            self.general_params_frame, menu_width=12,
            options=EXTRACT_TO_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.extract_to_menu, "extract_to"))
        self.prune_tiff_menu = ScrollMenu(
            self.general_params_frame, menu_width=12,
            options=NO_YES_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.prune_tiff_menu, "prune_tiff"))
        self.generate_mips_menu = ScrollMenu(
            self.general_params_frame, menu_width=12,
            options=NO_YES_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.generate_mips_menu, "mip_gen"))
        self.downres_box = tk.Spinbox(
            self.general_params_frame, from_=0, to=12, width=16,
            state="readonly", command=lambda *a, s=self:
            s.set_conversion_flag(self.downres_box, "downres"))


        self.platform_menu.tooltip_string = (
            "The platform to make the tag usable on.")
        self.format_menu.tooltip_string = (
            "The format to convert the bitmap to.")
        self.extract_to_menu.tooltip_string = (
            "The image format to extract the bitmap to.")
        self.prune_tiff_menu.tooltip_string = (
            "Prunes the uncompressed TIFF pixel data\n"
            "from all bitmaps to reduce their filesize.")
        self.generate_mips_menu.tooltip_string = (
            "Whether or not to generate all necessary mipmaps.")
        self.downres_box.tooltip_string = (
            "Number of times to cut the bitmaps\n"
            "width, height, and depth in half.")

        self.ay8_channel_src_menu = ScrollMenu(
            self.format_params_frame, menu_width=10,
            options=AY8_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.ay8_channel_src_menu, "mono_channel_to_keep"))
        self.ck_transparency_menu = ScrollMenu(
            self.format_params_frame, menu_width=10,
            options=NO_YES_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.ck_transparency_menu, "ck_trans"))
        self.swap_a8y8_menu = ScrollMenu(
            self.format_params_frame, menu_width=10,
            options=NO_YES_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.swap_a8y8_menu, "mono_swap"))
        self.multi_swap_menu = ScrollMenu(
            self.format_params_frame, menu_width=10,
            options=MULTI_SWAP_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.multi_swap_menu, "multi_swap"))
        self.swizzled_menu = ScrollMenu(
            self.format_params_frame, menu_width=10,
            options=NO_YES_OPTIONS, callback=lambda *a, s=self:
            s.set_conversion_flag(self.swizzled_menu, "swizzled"))
        self.alpha_bias_box = tk.Spinbox(
            self.format_params_frame, from_=0, to=255, width=14,
            state="readonly", repeatinterval=10, command=lambda *a, s=self:
            s.set_conversion_flag(self.alpha_bias_box, "alpha_bias"))


        self.ay8_channel_src_menu.tooltip_string = (
            "HUD meters converted to/from Xbox A8Y8 need to\n"
            "have their intensity and alpha channels swapped.\n"
            "Setting this will swap them when going to/from A8Y8.")
        self.ck_transparency_menu.tooltip_string = (
            "Whether to use color-key transparency when converting\n"
            "to p8-bump or DXT1. These formats support transparency\n"
            "where transparent pixels are also solid black in color.")
        self.swap_a8y8_menu.tooltip_string = (
            "Whether or not to swap the alpha and intensity\n"
            "channels when converting to or from A8Y8.")
        self.multi_swap_menu.tooltip_string = (
            "When converting multipurpose bitmaps to/from\n"
            "Xbox/PC, use this to swap the their color\n"
            "channels so they work on the other platform.")
        self.swizzled_menu.tooltip_string = (
            "Whether or not to swizzle the bitmap pixels.\n"
            "This does nothing to DXT1/3/5 bitmaps.\n"
            "Xbox bitmaps MUST be swizzled to work.")
        self.alpha_bias_box.tooltip_string = (
            "When converting to DXT1 with transparency, p8-bump,\n"
            "or A1R5G5B5, alpha values below this are rounded to\n"
            "black, while values at or above it round to white.")


        self.scan_button = tk.Button(self.buttons_frame, text="Scan directory",
                                     command=self.scan_pressed)
        self.convert_button = tk.Button(self.buttons_frame, text="Convert",
                                        command=self.convert_pressed)
        self.cancel_button = tk.Button(self.buttons_frame, text="Cancel",
                                       command=self.cancel_pressed)


        self.main_frame.pack(fill='both')
        self.tag_list_frame.pack(expand=True, fill='both')

        self.settings_frame.grid(sticky='news', row=0, column=0)
        self.bitmap_info_frame.grid(sticky='news', row=0, column=1)
        self.buttons_frame.grid(sticky='news', columnspan=2,
                                row=1, column=0, pady=3, padx=3)


        self.scan_button.pack(side='left', expand=True, fill='both', padx=3)
        self.convert_button.pack(side='left', expand=True, fill='both', padx=3)
        self.cancel_button.pack(side='left', expand=True, fill='both', padx=3)

        self.scan_dir_frame.pack(expand=True, fill='x')
        self.data_dir_frame.pack(expand=True, fill='x')
        self.log_file_frame.pack(expand=True, fill='x')
        self.global_params_frame.pack(expand=True, fill='x')
        self.params_frame.pack(expand=True, fill='x')

        self.general_params_frame.pack(side='left', expand=True, fill='both')
        self.format_params_frame.pack(side='left', expand=True, fill='both')

        self.log_file_entry.pack(side='left', expand=True, fill='x')
        self.log_file_browse_button.pack(side='left')

        self.scan_dir_entry.pack(side='left', expand=True, fill='x')
        self.scan_dir_browse_button.pack(side='left')

        self.data_dir_entry.pack(side='left', expand=True, fill='x')
        self.data_dir_browse_button.pack(side='left')


        self.read_only_cbutton.grid(row=0, column=0, sticky='w')
        self.backup_tags_cbutton.grid(row=0, column=1, sticky='w')
        self.open_log_cbutton.grid(row=0, column=2, sticky='w')
        self.use_stubbs_p8_cbutton.grid(row=0, column=3, sticky='w')

        i = 0
        widgets = (self.platform_menu, self.format_menu, self.extract_to_menu,
                   self.prune_tiff_menu, self.generate_mips_menu, self.downres_box)
        for name in ("Platform", "Format", "Extract to",
                     "Prune TIFF data", "Generate mipmaps", "Downres. level"):
            w = widgets[i]
            lbl = tk.Label(self.general_params_frame, text=name)
            lbl.grid(row=i, column=0, sticky='w')
            w.grid(row=i, column=1, sticky='e')
            try:
                if w.tooltip_string:
                    lbl.tooltip_string = w.tooltip_string
                    lbl.config(text=lbl.config()["text"][-1] + " �")
            except AttributeError:
                pass
            i += 1

        i = 0
        widgets = (self.ay8_channel_src_menu, self.ck_transparency_menu,
                   self.swap_a8y8_menu, self.multi_swap_menu,
                   self.swizzled_menu, self.alpha_bias_box)
        for name in ("AY8 channel source", "Use CK transparency",
                     "Swap A8Y8", "Multi. swap",
                     "Swizzled", "Alpha bias"):
            w = widgets[i]
            lbl = tk.Label(self.format_params_frame, text=name)
            lbl.grid(row=i, column=0, sticky='w')
            w.grid(row=i, column=1, sticky='e')
            try:
                if w.tooltip_string:
                    lbl.tooltip_string = w.tooltip_string
                    lbl.config(text=lbl.config()["text"][-1] + " �")
            except AttributeError:
                pass
            i += 1


        self.curr_bitmap_label_0.pack(side='left', fill='x')
        self.curr_bitmap_spinbox.pack(side='left', fill='x', expand=True)
        self.curr_bitmap_label_1.pack(side='left', fill='x')
        self.max_bitmap_entry.pack(side='left', fill='x', expand=True)

        self.bitmap_index_frame.grid(row=0, column=0, sticky='we',
                                     columnspan=2, pady=(5, 10))
        i = 1
        widgets = (self.curr_type_menu, self.curr_format_menu,
                   self.curr_swizzled_menu,
                   self.curr_platform_menu, self.curr_has_tiff_menu,
                   self.curr_width_entry, self.curr_height_entry,
                   self.curr_depth_entry, self.curr_mip_entry)
        for name in ("Type", "Format", "Swizzled", "Platform", "Has TIFF",
                     "Width", "Height", "Depth", "Mipmaps"):
            w = widgets[i - 1]
            lbl = tk.Label(self.bitmap_info_frame, text=name)
            lbl.grid(row=i, column=0, sticky='w', pady=3)
            w.grid(row=i, column=1, sticky='e')
            try:
                if w.tooltip_string:
                    lbl.tooltip_string = w.tooltip_string
                    lbl.config(text=lbl.config()["text"][-1] + " �")
            except AttributeError:
                pass
            i += 1

        widgets = self.children.values()
        while widgets:
            next_widgets = []
            for w in widgets:
                try:
                    if isinstance(w, ScrollMenu):
                        s = w.tooltip_string
                        w.sel_label.tooltip_string = s
                        w.button_frame.tooltip_string = s
                        w.arrow_button.tooltip_string = s
                    else:
                        next_widgets.extend(w.children.values())
                except Exception:
                    print(format_exc())
            widgets = next_widgets

        self.buttons = (self.scan_dir_browse_button, self.scan_button,
                        self.log_file_browse_button, self.convert_button)
        self.checkbuttons = (self.read_only_cbutton, self.backup_tags_cbutton,
                             self.open_log_cbutton, self.use_stubbs_p8_cbutton)
        self.spinboxes = (self.downres_box, self.alpha_bias_box)
        self.menus = (self.platform_menu, self.format_menu,
                      self.extract_to_menu, self.prune_tiff_menu,
                      self.multi_swap_menu, self.generate_mips_menu,
                      self.ay8_channel_src_menu, self.ck_transparency_menu,
                      self.swap_a8y8_menu, self.swizzled_menu)

        self.apply_style()
        self.populate_bitmap_info()
        self.populate_settings()

        if self.app_root is not self:
            self.transient(self.app_root)

    def update_all_path_colors(self):
        update_color = self.tag_list_frame.update_path_listbox_entry_color
        for i in range(self.tag_list_frame.path_listbox.size()):
            update_color(i)

    def destroy(self):
        try:
            self.app_root.tool_windows.pop(self.window_name, None)
        except AttributeError:
            pass
        super(type(self), self).destroy()

    def apply_style(self, seen=None):
        BinillaWidget.apply_style(self, seen)
        self.update()
        w, h = self.winfo_reqwidth(), self.winfo_reqheight()
        self.geometry("%sx%s" % (w, h))
        self.minsize(width=w, height=h)

    def show_log_in_text_editor(self):
        Thread(target=self._show_log_in_text_editor, daemon=True).start()

    def _show_log_in_text_editor(self):
        try:
            log_path = self.log_file_path.get()
            if not os.path.exists(log_path):
                return

            do_subprocess(TEXT_EDITOR_NAME, (), (log_path, ),
                          proc_controller=ProcController(abandon=True),
                          stdout=None, stderr=None, stdin=None)
        except Exception:
            print(format_exc())

    def set_conversion_flag(self, widget, flag_name):
        new_value = -1
        if isinstance(widget, ScrollMenu):
            new_value = widget.sel_index
        elif isinstance(widget, tk.Spinbox):
            new_value = int(widget.get())

        if new_value < 0:
            return

        conv_flags = self.conversion_flags
        path_listbox = self.tag_list_frame.path_listbox
        update_color = self.tag_list_frame.update_path_listbox_entry_color
        for i in path_listbox.curselection():
            fp = path_listbox.get(i)

            bitm_tag_info = self.bitmap_tag_infos.get(fp)
            if flag_name == "new_format" and bitm_tag_info:
                fmt = PARAM_FORMAT_TO_FORMAT[new_value]
                if bitm_tag_info.type == 1 and fmt in (14, 15, 16):
                    print("Cannot convert 3D textures to DXT.")
                    continue
                elif bitm_tag_info.type == 2 and fmt == 17:
                    print("Cannot convert cubemaps textures to P8.")
                    continue

            if conv_flags.get(fp):
                setattr(conv_flags[fp], flag_name, new_value)

            update_color(i)

    def initialize_conversion_flags(self):
        data_dir = self.data_dir_path.get()
        for fp, info in self.bitmap_tag_infos.items():
            if not info:
                continue

            self.conversion_flags[fp] = flags = ConversionFlags()

            flags.platform = info.platform
            flags.swizzled = info.swizzled
            flags.extract_path = os.path.splitext(
                os.path.join(data_dir, fp))[0]

    def scan_pressed(self):
        if self._processing:
            return

        new_tags_dir = self.scan_dir_path.get()
        if not os.path.exists(new_tags_dir):
            print("The specified directory to scan does not exist.")
            return

        self.conversion_flags = {}
        self.bitmap_tag_infos = {}
        self.bitmap_display_windows = {}
        self.loaded_tags_dir = new_tags_dir

        try: self.scan_thread.join()
        except Exception: pass
        self.disable_settings()
        self.scan_thread = Thread(target=self._scan)
        self.scan_thread.daemon = True
        self.scan_thread.start()

    def _scan(self):
        self._processing = True
        try:
            print("Locating bitmaps...")

            s_time = time()
            c_time = s_time
            p_int = self.print_interval

            seen_files = set()  # keep track of all absolute filepaths seen
            scan_dir = self.loaded_tags_dir
            for root, _, files in os.walk(scan_dir):
                rel_root = Path(os.path.relpath(root, scan_dir))
                root = Path(root)

                for filename in files:
                    if Path(filename).suffix.lower() != ".bitmap":
                        continue

                    # We use an absolute path here because we later store the
                    # bitmap tag info as hash keys. This is just so that we
                    # don't do any double conversions.
                    try:
                        filepath = root.joinpath(filename).resolve()
                        if filepath in seen_files:
                            continue

                        seen_files.add(filepath)
                    except FileNotFoundError:
                        continue

                    rel_filepath = str(rel_root.joinpath(filename))
                    if time() - c_time > p_int:
                        c_time = time()
                        print('    ' + rel_filepath)
                        if self.app_root:
                            self.app_root.update_idletasks()

                    if self._cancel_processing:
                        print('Bitmap scanning cancelled.\n')
                        self.after(0, self.enable_settings)
                        return

                    try:
                        bitm_tag = self.bitm_def.build(filepath=filepath)
                    except Exception:
                        print(format_exc())
                        bitm_tag = None

                    if not bitm_tag:
                        print("Could not load: %s" % filepath)
                        continue

                    self.bitmap_tag_infos[rel_filepath] = BitmapTagInfo(bitm_tag)

            print("    Finished in %s seconds." % int(time() - s_time))
        except Exception:
            print(format_exc())

        self.initialize_conversion_flags()
        self.tag_list_frame.build_tag_sort_mappings()
        self.tag_list_frame.display_sorted_tags()
        self.after(0, self.populate_bitmap_info)
        self.after(0, self.populate_settings)
        self.after(0, self.enable_settings)
        self._processing = False

    def convert_pressed(self):
        if self._processing or not self.bitmap_tag_infos:
            return

        try: self.convert_thread.join()
        except Exception: pass
        self.disable_settings()
        self.convert_thread = Thread(target=self._convert)
        self.convert_thread.daemon = True
        self.convert_thread.start()

    def _convert(self):
        self._processing = True
        s_time = time()
        c_time = s_time

        if self.read_only.get():
            print("Creating log...")
            try:
                if self.make_log() and self.open_log.get():
                    self.show_log_in_text_editor()
            except Exception:
                print(format_exc())
                print("Could not create log")

        else:
            print("Converting bitmaps...")
            tags_dir = self.loaded_tags_dir

            for fp in sorted(self.bitmap_tag_infos):
                try:
                    if self._cancel_processing:
                        print("Conversion cancelled by user.")
                        break

                    bitmap_info = self.bitmap_tag_infos[fp]
                    conv_flags = self.conversion_flags[fp]
                    pruning = conv_flags.prune_tiff
                    extracting = conv_flags.extract_to != 0
                    converting = get_will_be_converted(conv_flags, bitmap_info)
                    if pruning or converting or extracting:
                        tag = self.bitm_def.build(filepath=os.path.join(tags_dir, fp))
                        if pruning:
                            tag.data.tagdata.compressed_color_plate_data.data = bytearray()

                        if converting or extracting:
                            convert_bitmap_tag(tag, conv_flags, bitmap_info,
                                               use_stubbs_p8=self.use_stubbs_p8.get())

                        if converting or pruning:
                            tag.serialize(temp=False, calc_pointers=False,
                                          backup=self.backup_tags.get())

                        self.bitmap_tag_infos.pop(fp, None)
                        self.conversion_flags.pop(fp, None)
                        self.bitmap_display_windows.pop(fp, None)

                        del tag
                        gc.collect()
                except Exception:
                    print(format_exc())
                    print("Could not convert: %s" % fp)

        print("    Finished in %s seconds." % int(time() - s_time))

        self._processing = self._cancel_processing = False
        self.after(0, self.enable_settings)
        self.after(0, self.tag_list_frame.display_sorted_tags)

    def cancel_pressed(self):
        if self._processing:
            self._cancel_processing = True

    def enable_settings(self):
        self._enable_disable_settings(False)

    def disable_settings(self):
        self._enable_disable_settings(True)

    def _enable_disable_settings(self, disable):
        if disable != self._settings_enabled:
            return

        new_state = tk.DISABLED if disable else tk.NORMAL
        for w in self.checkbuttons + self.buttons:
            w.config(state=new_state)

        self._settings_enabled = not disable
        if disable:
            self._settings_enabled = False
            new_state = "readonly"

        for w in self.spinboxes:
            w.config(state=new_state)

        for w in self.menus:
            if disable:
                w.disable()
            else:
                w.enable()

    def populate_settings(self):
        if self._populating_settings or self._processing:
            return

        self._populating_settings = True
        settings_enabled = self._settings_enabled
        try:
            self.enable_settings()
            menus = (self.platform_menu, self.format_menu,
                     self.extract_to_menu, self.prune_tiff_menu,
                     self.multi_swap_menu, self.generate_mips_menu,
                     self.ay8_channel_src_menu, self.ck_transparency_menu,
                     self.swap_a8y8_menu, self.swizzled_menu)
            for w in menus:
                w.sel_index = -1

            for w in (self.downres_box, self.alpha_bias_box):
                w.delete(0, tk.END)

            conv_flags = self.conversion_flags
            tag_paths = self.tag_list_frame.selected_paths
            comb_flags = None
            if not tag_paths:
                self._populating_settings = False
                return

            for tag_path in tag_paths:
                comb_flags = deepcopy(conv_flags.get(tag_path))
                break

            if not comb_flags:
                comb_flags = ConversionFlags()
                print("Could not locate conversion settings for this/these tags.")

            for fp in tag_paths:
                flags = conv_flags.get(fp)
                if not flags:
                    continue

                for name in ("platform", "multi_swap", "prune_tiff",
                             "mono_channel_to_keep", "extract_to", "new_format",
                             "swizzled", "mono_swap", "ck_trans", "mip_gen",
                             "downres", "alpha_bias"):
                    if getattr(flags, name) != getattr(comb_flags, name):
                        setattr(comb_flags, name, -1)

            self.platform_menu.sel_index = comb_flags.platform
            self.multi_swap_menu.sel_index = comb_flags.multi_swap
            self.ay8_channel_src_menu.sel_index = comb_flags.mono_channel_to_keep
            self.extract_to_menu.sel_index = comb_flags.extract_to
            self.prune_tiff_menu.sel_index = comb_flags.prune_tiff
            self.format_menu.sel_index = comb_flags.new_format
            self.swizzled_menu.sel_index = comb_flags.swizzled
            self.swap_a8y8_menu.sel_index = comb_flags.mono_swap
            self.ck_transparency_menu.sel_index = comb_flags.ck_trans
            self.generate_mips_menu.sel_index = comb_flags.mip_gen

            for w in menus:
                if w.sel_index < 0:
                    w.update_label("<mixed values>")

            for name, w in (("downres", self.downres_box),
                            ("alpha_bias", self.alpha_bias_box)):
                w.delete(0, tk.END)
                val = getattr(comb_flags, name)
                w.insert(0, str(val) if val >= 0 else "<mixed values>")

            if not settings_enabled:
                self.disable_settings()
        except Exception:
            print(format_exc())
        self._populating_settings = False

    def populate_bitmap_info(self):
        if self._populating_bitmap_info or self._processing:
            return

        self._populating_bitmap_info = True
        try:
            try:
                i = int(self.curr_bitmap_spinbox.get())
            except ValueError:
                i = 0

            for w in (self.curr_width_entry, self.curr_height_entry,
                      self.curr_depth_entry, self.curr_mip_entry,
                      self.curr_bitmap_spinbox, self.max_bitmap_entry):
                w.config(state=tk.NORMAL)

            for w in (self.curr_type_menu, self.curr_format_menu,
                      self.curr_swizzled_menu,
                      self.curr_platform_menu, self.curr_has_tiff_menu):
                w.sel_index = -1

            for w in (self.curr_width_entry, self.curr_height_entry,
                      self.curr_depth_entry, self.curr_mip_entry,
                      self.max_bitmap_entry, self.curr_bitmap_spinbox):
                w.delete(0, tk.END)

            tag_paths = self.tag_list_frame.selected_paths
            if len(tag_paths) == 1:
                for tag_path in tag_paths:
                    bitm_tag_info = self.bitmap_tag_infos.get(tag_path)

                if bitm_tag_info:
                    bitm_ct = len(bitm_tag_info.bitmap_infos)

                    if i >= bitm_ct:
                        i = 0

                    if bitm_ct:
                        self.curr_bitmap_spinbox.config(to=bitm_ct - 1)
                        self.curr_bitmap_spinbox.insert(tk.END, str(i))
                        self.max_bitmap_entry.insert(tk.END, str(bitm_ct - 1))
                    else:
                        self.curr_bitmap_spinbox.config(to=1)
                        self.curr_bitmap_spinbox.insert(tk.END, "")
                        self.max_bitmap_entry.insert(tk.END, "")

                    if i < bitm_ct:
                        bitm_info = bitm_tag_info.bitmap_infos[i]

                        self.curr_type_menu.sel_index = bitm_info.type
                        self.curr_format_menu.sel_index = bitm_info.format
                        self.curr_swizzled_menu.sel_index = bitm_info.swizzled
                        self.curr_platform_menu.sel_index = bitm_tag_info.platform
                        self.curr_has_tiff_menu.sel_index = bitm_tag_info.tiff_data_size > 0

                        self.curr_width_entry.insert(tk.END, str(bitm_info.width))
                        self.curr_height_entry.insert(tk.END, str(bitm_info.height))
                        self.curr_depth_entry.insert(tk.END, str(bitm_info.depth))
                        self.curr_mip_entry.insert(tk.END, str(bitm_info.mipmaps))
                else:
                    print("Could not locate bitmap info for: %s" % tag_path)

        except Exception:
            print(format_exc())

        for w in (self.curr_width_entry, self.curr_height_entry,
                  self.curr_depth_entry, self.curr_mip_entry):
            w.config(state=tk.DISABLED)

        for w in (self.curr_bitmap_spinbox, self.max_bitmap_entry):
            w.config(state="readonly")

        self._populating_bitmap_info = False

    def scan_dir_browse(self):
        if self._processing:
            return

        load_dir = self.scan_dir_path.get()
        if not load_dir:
            load_dir = self.app_root.last_load_dir
        dirpath = askdirectory(initialdir=load_dir,
                               parent=self, title="Select directory to scan")
        if not dirpath:
            return

        dirpath = Path(dirpath)

        self.scan_dir_path.set(dirpath)
        if self.app_root:
            self.app_root.last_load_dir = dirpath.parent

    def data_dir_browse(self):
        if self._processing:
            return

        load_dir = self.data_dir_path.get()
        if not load_dir:
            load_dir = self.app_root.last_load_dir
        dirpath = askdirectory(initialdir=self.data_dir_path.get(),
                               parent=self, title="Select directory to scan")
        if not dirpath:
            return

        dirpath = Path(dirpath)

        curr_data_dir = self.data_dir_path.get()
        for flags in self.conversion_flags.values():
            if not flags:
                continue

            flags.extract_path = dirpath.join_path(
                Path(flags.extract_path).relative_to(curr_data_dir)
            )

        self.data_dir_path.set(dirpath)

    def log_browse(self):
        if self._processing:
            return

        load_dir = os.path.dirname(self.log_file_path.get())
        if not load_dir:
            load_dir = self.app_root.last_load_dir
        fp = asksaveasfilename(
            initialdir=load_dir, title="Save scan log to...", parent=self,
            filetypes=(("bitmap optimizer log", "*.log"), ('All', '*')))

        if not fp:
            return

        fp = Path(fp)
        if not fp.suffix:
            fp = fp.with_suffix('.log')

        self.log_file_path.set(fp)
        if self.app_root:
            self.app_root.last_load_dir = Path(self.log_file_path.get()).parent

    def get_will_be_processed(self, tag_path):
        info = self.bitmap_tag_infos.get(tag_path)
        if self.read_only.get() or (not info or not info.bitmap_infos):
            return False

        if self.conversion_flags[tag_path].prune_tiff and info.tiff_data_size:
            return True
        elif self.conversion_flags[tag_path].extract_to != 0:
            return True
        return get_will_be_converted(self.conversion_flags[tag_path],
                                     self.bitmap_tag_infos[tag_path])

    def make_log(self):
        attempts = 0
        success = True

        while attempts < 2:
            log_file_path = self.log_file_path.get()
            try:
                if not os.path.exists(os.path.dirname(log_file_path)):
                    log_file_path = None

                if not log_file_path.endswith(".log"):
                    log_file_path += ".log"
            except Exception:
                log_file_path = None

            attempts += 1
            if attempts < 2 and not log_file_path:
                self._processing = False
                self.log_file_path.set("")
                self.log_browse()

        logstr = "Mozzarilla Bitmap Converter tagset log:\n\n"

        total_size = 0
        tiff_data_size = 0
        for filename, info in self.bitmap_tag_infos.items():
            total_size += os.path.getsize(os.path.normpath(
                os.path.join(self.loaded_tags_dir, filename)))
            tiff_data_size += info.tiff_data_size

        logstr += "%s bitmaps total\n%sKB of bitmap data\n%sKB of TIFF data" % (
            len(self.bitmap_tag_infos), total_size // 1024, tiff_data_size // 1024)

        tag_counts = [0, 0, 0]
        tag_header_strs = ("2D Textures", "3D Textures", "Cubemaps")
        base_str = "Bitmap %s\t--- WxHxD: %sx%sx%s\t--- Mipmaps: %s\n"

        tag_info_strs = {}
        formatted_strs = {}

        for typ in range(HALO_1_TYPE_COUNT):
            formatted_strs[typ] = [''] * HALO_1_FORMAT_COUNT
            tag_info_strs[typ]  = [''] * HALO_1_FORMAT_COUNT

            for fmt in range(HALO_1_FORMAT_COUNT):
                if "?" not in BITMAP_FORMATS[fmt]:
                    formatted_strs[typ][fmt] = "\n\n\t%s" % BITMAP_FORMATS[fmt]
                    tag_info_strs[typ][fmt] = {}

        for filename, info in self.bitmap_tag_infos.items():
            fp = os.path.normpath(
                os.path.join(self.loaded_tags_dir, filename))
            filesize = (os.path.getsize(fp) - info.tiff_data_size) // 1024
            tagstr = ("\n\t\t" + fp +
                      "\n\t\t\tCompiled tag size\t= %sKB\n" %
                      ("less than 1" if filesize <= 0 else str(filesize)))

            if info.tiff_data_size > 0:
                tagstr += "\t\t\tTIFF data size\t= %sKB\n" % (
                    "less than 1" if info.tiff_data_size < 1024
                    else info.tiff_data_size // 1024)

            i = 0
            for bitm_info in info.bitmap_infos:
                tagstr += ("\t\t\t" + base_str %
                           (i, bitm_info.width, bitm_info.height,
                            bitm_info.depth, bitm_info.mipmaps))
                i += 1

            tag_strs = tag_info_strs[info.type][info.format]
            tag_strs.setdefault(filesize, [])
            tag_strs[filesize].append(tagstr)

        for typ in range(HALO_1_TYPE_COUNT):
            for fmt in VALID_FORMAT_ENUMS:
                for size in reversed(sorted(tag_info_strs[typ][fmt])):
                    for tagstr in tag_info_strs[typ][fmt][size]:
                        tag_counts[typ] += 1
                        formatted_strs[typ][fmt] += tagstr

        for typ in range(HALO_1_TYPE_COUNT):
            logstr += "\n\n%s:\n\tCount = %s%s" % (
                tag_header_strs[typ], tag_counts[typ],
                ''.join(formatted_strs[typ]))

        if log_file_path:
            try:
                with open(log_file_path, "w") as f:
                    f.write(logstr)
            except Exception:
                print(format_exc())
                success = False
        else:
            print(logstr)
            success = False

        return success


class BitmapConverterList(tk.Frame, BinillaWidget, HaloBitmapDisplayBase):
    listboxes = ()
    reverse_listbox = False
    toggle_to = True
    sort_method = 'path'

    displayed_paths = ()
    type_format_map = ()
    selected_paths = ()

    _populating = False

    def __init__(self, master, **options):
        BinillaWidget.__init__(self, master, **options)
        tk.Frame.__init__(self, master, **options)

        self.formats_shown = [True] * HALO_1_FORMAT_COUNT
        self.types_shown   = [True] * HALO_1_TYPE_COUNT
        self.displayed_paths = []
        self.type_format_map = []
        self.selected_paths = set()
        self.build_tag_sort_mappings()

        self.sort_menu = tk.Menu(self, tearoff=False)
        self.types_menu = tk.Menu(self, tearoff=False)
        self.formats_menu = tk.Menu(self, tearoff=False)

        self.sort_menu.add_command(
            label="Toggle all to %s" % BITMAP_PLATFORMS[int(self.toggle_to)],
            command=self.toggle_all)
        self.sort_menu.add_command(
            label="Invert selection", command=self.invert_selection)
        self.sort_menu.add_separator()
        self.sort_menu.add_command(
            label="Sort Ascending", command=lambda:
            self.display_sorted_tags(None, False))
        self.sort_menu.add_command(
            label="Sort Descending", command=lambda:
            self.display_sorted_tags(None, True))
        self.sort_menu.add_separator()

        self.sort_menu.add_command(
            label="Sort by Filepath", command=lambda:
            self.display_sorted_tags('path'))
        self.sort_menu.add_command(
            label="Sort by Bitmap data size", command=lambda:
            self.display_sorted_tags('size'))
        self.sort_menu.add_command(
            label="Sort by Bitmap format", command=lambda:
            self.display_sorted_tags('format'))
        self.sort_menu.add_command(
            label="Sort by Bitmap type", command=lambda:
            self.display_sorted_tags('type'))

        self.types_menu.add_command(
            label="Toggle all", command=lambda:
            self.toggle_types_allowed(0, -1))
        for typ in range(len(BITMAP_TYPES)):
            self.types_menu.add_command(
                label=BITMAP_TYPES[typ] + u' \u2713',
                command=lambda t=typ: self.toggle_types_allowed(t + 1, t))

        self.formats_menu.add_command(
            label="Toggle all", command=lambda:
            self.toggle_formats_allowed(0, -1))
        i = 1
        for fmt in VALID_FORMAT_ENUMS:
            self.formats_menu.add_command(
                label=BITMAP_FORMATS[fmt] + u' \u2713',
                command=lambda i=i, f=fmt:self.toggle_formats_allowed(i, f))
            i += 1

        self.vsb = tk.Scrollbar(self, orient="vertical")
        self.hsb = tk.Scrollbar(self, orient="horizontal")
        self.listboxes = []
        self.listboxes.append(
            tk.Listbox(self, height=5, exportselection=False,
                       yscrollcommand=self._path_scrolled,
                       xscrollcommand=self.hsb.set))
        self.listboxes.append(
            tk.Listbox(self, width=8, height=5, exportselection=False,
                       yscrollcommand=self._size_scrolled))
        self.listboxes.append(
            tk.Listbox(self, width=11, height=5, exportselection=False,
                       yscrollcommand=self._format_scrolled))
        self.listboxes.append(
            tk.Listbox(self, width=6, height=5, exportselection=False,
                       yscrollcommand=self._type_scrolled))

        self.path_listbox.bind("<Button-3>", lambda e, m=self.sort_menu:
                               self.post_rightclick_menu(e, m))
        self.size_listbox.bind("<Button-3>", lambda e, m=self.sort_menu:
                               self.post_rightclick_menu(e, m))
        self.format_listbox.bind("<Button-3>", lambda e, m=self.formats_menu:
                                 self.post_rightclick_menu(e, m))
        self.type_listbox.bind("<Button-3>", lambda e, m=self.types_menu:
                               self.post_rightclick_menu(e, m))

        self.hsb.config(command=self.path_listbox.xview)
        self.vsb.config(command=self._scroll_all_yviews)
        self.listboxes[0].bind('<<ListboxSelect>>', self.set_selected_tags_list)

        for i in range(len(self.listboxes)):
            self.listboxes[i].config(selectmode=tk.EXTENDED, highlightthickness=0)
            self.listboxes[i].bind('<Double-Button-1>', lambda e, idx=i:
                                   self.display_selected_tag(idx))
            self.listboxes[i].bind('<Return>', lambda e, idx=i:
                                   self.display_selected_tag(idx))
            if i != 0:
                self.listboxes[i].bind('<<ListboxSelect>>', lambda e, idx=i:
                                       self.select_path_listbox(idx))

        self.hsb.pack(side="bottom", fill="x")
        self.vsb.pack(side="right",  fill="y")
        self.path_listbox.pack(side="left", fill="both", expand=True)
        for listbox in self.listboxes[1: ]:
            listbox.pack(side="left", fill="both")

        self.apply_style()

    def get_p8_palette(self, tag=None):
        if not self.master.use_stubbs_p8 or not self.master.use_stubbs_p8.get():
            return HALO_P8_PALETTE
        return STUBBS_P8_PALETTE

    @property
    def path_listbox(self): return self.listboxes[0]
    @property
    def size_listbox(self): return self.listboxes[1]
    @property
    def format_listbox(self): return self.listboxes[2]
    @property
    def type_listbox(self): return self.listboxes[3]

    def apply_style(self, seen=None):
        BinillaWidget.apply_style(self, seen)
        self.populate_tag_list_boxes()

    def post_rightclick_menu(self, event, menu):
        self.update_sort_menu()
        menu.post(event.x_root, event.y_root)

    def reset_listboxes(self):
        self.selected_paths = set()
        self.displayed_paths = []
        for listbox in self.listboxes:
            listbox.delete(0, tk.END)

    def set_selected_tags_list(self, event=None):
        selected = set(self.path_listbox.curselection())
        if len(selected) == 1:
            self.selected_paths.clear()
        else:
            for i in range(self.path_listbox.size()):
                if i not in selected:
                    fp = self.path_listbox.get(i)
                    if fp in self.selected_paths:
                        self.selected_paths.remove(fp)

        for i in selected:
            self.selected_paths.add(self.path_listbox.get(i))

        self.master.populate_bitmap_info()
        self.master.populate_settings()

    def toggle_all(self):
        for flags in self.master.conversion_flags.values():
            flags.swizzled = flags.platform = self.toggle_to

        self.toggle_to = not self.toggle_to
        self.display_sorted_tags()
        self.master.populate_settings()

    def update_sort_menu(self):
        sort_menu_strs = []
        for i in range(10):
            try:
                label = self.sort_menu.entryconfig(i)["label"][-1]
            except Exception:
                label = ""

            sort_menu_strs.append(label.rstrip(u' \u2713'))

        sort_menu_strs[0] = "Toggle all to %s" % BITMAP_PLATFORMS[int(self.toggle_to)]

        if self.reverse_listbox:
            sort_menu_strs[4] += u' \u2713'
        else:
            sort_menu_strs[3] += u' \u2713'

        if self.sort_method == 'path':
            sort_menu_strs[6] += u' \u2713'
        elif self.sort_method == 'size':
            sort_menu_strs[7] += u' \u2713'
        elif self.sort_method == 'format':
            sort_menu_strs[8] += u' \u2713'
        elif self.sort_method == 'type':
            sort_menu_strs[9] += u' \u2713'

        for i in range(len(sort_menu_strs)):
            if sort_menu_strs[i]:
                self.sort_menu.entryconfig(i, label=sort_menu_strs[i])

    def invert_selection(self):
        for i in range(self.path_listbox.size()):
            fp = self.path_listbox.get(i)
            if fp in self.selected_paths:
                self.selected_paths.remove(fp)
            else:
                self.selected_paths.add(fp)

        self.synchronize_selection()
        self.set_selected_tags_list()

    def synchronize_selection(self):
        self.path_listbox.selection_clear(0, tk.END)
        for i in range(self.path_listbox.size()):
            if self.path_listbox.get(i) in self.selected_paths:
                self.path_listbox.selection_set(i)

    def display_selected_tag(self, src_listbox_index=0):
        self.select_path_listbox(src_listbox_index)
        if len(self.selected_paths) != 1 or not self.master.loaded_tags_dir:
            return

        for tag_path in self.selected_paths:
            break

        display_frame = self.master.bitmap_display_windows.get(tag_path)
        if display_frame is None or display_frame() is None:
            tags_dir = self.master.loaded_tags_dir
            tag = self.master.bitm_def.build(filepath=os.path.join(tags_dir, tag_path))

            if not tag:
                print("Could not load the tag: %s" % tag_path)
                return

            # the bitmap display frame requires tags_dir to see
            # it as an actual tag rather than a meta tag
            tag.tags_dir = tags_dir

            w = tk.Toplevel(self.master)
            display_frame = weakref.ref(HaloBitmapDisplayFrame(w, tag))

            display_frame().change_textures(self.get_textures(tag))
            display_frame().pack(expand=True, fill="both")

            w.title("Preview: %s" % tag_path)
            w.transient(self.master)

            self.master.bitmap_display_windows[tag_path] = display_frame

        w.update_idletasks()
        display_frame().focus_set()
        self.master.place_window_relative(w)

    def select_path_listbox(self, src_listbox_index=0):
        src_listbox = self.listboxes[src_listbox_index]

        if len(src_listbox.curselection()) > 0:
            self.path_listbox.selection_set(src_listbox.curselection()[0])
            self.set_selected_tags_list()

        if src_listbox_index > 0:
            src_listbox.selection_clear(0, tk.END)

        self.master.populate_bitmap_info()
        self.master.populate_settings()

    def toggle_types_allowed(self, menu_idx, typ):
        if typ == -1:
            for typ in range(HALO_1_TYPE_COUNT):
                self.types_shown[typ] = not self.types_shown[typ]
                typ_str = BITMAP_TYPES[typ]
                if self.types_shown[typ]: typ_str += u' \u2713'

                self.types_menu.entryconfig(typ + 1, label=typ_str)
            self.display_sorted_tags()
            return

        typ_str = BITMAP_TYPES[typ]
        self.types_shown[typ] = not self.types_shown[typ]
        if self.types_shown[typ]: typ_str += u' \u2713'

        self.types_menu.entryconfig(menu_idx, label=typ_str)
        self.display_sorted_tags()

    def toggle_formats_allowed(self, menu_idx, fmt):
        if fmt == -1:
            i = 1 if menu_idx == 0 else 0
            for fmt in range(HALO_1_FORMAT_COUNT):
                if fmt in VALID_FORMAT_ENUMS:
                    self.formats_shown[fmt] = not self.formats_shown[fmt]
                    fmt_str = BITMAP_FORMATS[fmt]
                    if self.formats_shown[fmt]: fmt_str += u' \u2713'

                    self.formats_menu.entryconfig(i, label=fmt_str)
                    i += 1
            self.display_sorted_tags()
            return

        fmt_str = BITMAP_FORMATS[fmt]
        self.formats_shown[fmt] = not self.formats_shown[fmt]
        if self.formats_shown[fmt]: fmt_str += u' \u2713'

        self.formats_menu.entryconfig(menu_idx, label=fmt_str)
        self.display_sorted_tags()

    def build_tag_sort_mappings(self):
        self.selected_paths = set()
        self.displayed_paths = []
        self.type_format_map = []
        self.size_map = {}

        for typ in range(HALO_1_TYPE_COUNT):
            self.type_format_map.append([])
            for fmt in range(HALO_1_FORMAT_COUNT):
                self.type_format_map[typ].append([])

        remove = set()
        for fp, info in self.master.bitmap_tag_infos.items():
            if not info.pixel_data_size in self.size_map:
                self.size_map[info.pixel_data_size] = []

            try:
                self.type_format_map[info.type][info.format]
                self.size_map[info.pixel_data_size]

                self.type_format_map[info.type][info.format].append(fp)
                self.size_map[info.pixel_data_size].append(fp)
            except IndexError:
                remove.add(fp)

        for fp in remove:
            self.master.conversion_flags.pop(fp, None)
            self.master.bitmap_tag_infos.pop(fp, None)

    def display_sorted_tags(self, sort_by=None, reverse=None):
        if sort_by is None:
            sort_by = self.sort_method

        if reverse is not None:
            self.reverse_listbox = reverse

        self.sort_displayed_tags(sort_by)
        self.after(0, self.populate_tag_list_boxes)

    def sort_displayed_tags(self, sort_by):
        self.displayed_paths = displayed_paths = []
        if not self.master.bitmap_tag_infos:
            return

        if sort_by == 'path':
            for path in sorted(self.master.bitmap_tag_infos.keys()):
                info = self.master.bitmap_tag_infos[path]
                if (self.formats_shown[info.format] and
                    self.types_shown[info.type]):
                    displayed_paths.append(path)

        elif sort_by == 'size':
            for tagsize in sorted(self.size_map):
                for path in self.size_map[tagsize]:
                    info = self.master.bitmap_tag_infos[path]
                    if (self.formats_shown[info.format] and
                        self.types_shown[info.type]):
                        displayed_paths.append(path)

        elif sort_by == 'format':
            for fmt in range(len(self.formats_shown)):
                for typ in range(len(self.types_shown)):
                    if self.formats_shown[fmt] and self.types_shown[typ]:
                        displayed_paths.extend(self.type_format_map[typ][fmt])

        elif sort_by == 'type':
            for typ in range(len(self.types_shown)):
                for fmt in range(len(self.formats_shown)):
                    if self.formats_shown[fmt] and self.types_shown[typ]:
                        displayed_paths.extend(self.type_format_map[typ][fmt])
        else:
            for path in self.master.bitmap_tag_infos.keys():
                info = self.master.bitmap_tag_infos[path]
                if (self.formats_shown[info.format] and
                    self.types_shown[info.type]):
                    displayed_paths.append(path)

        self.sort_method = sort_by
        if self.reverse_listbox:
            displayed_paths.reverse()

    def populate_tag_list_boxes(self):
        if self._populating:
            return

        self._populating = True
        try:
            for listbox in self.listboxes:
                listbox.delete(0, tk.END)

            for fp in self.displayed_paths:
                try:
                    info = self.master.bitmap_tag_infos[fp]
                except KeyError:
                    continue

                size = info.pixel_data_size
                if size < 1024:
                    size_str = str(size) + "  B"
                elif size < 1024**2:
                    size_str = str((size + 512) // 1024) + "  KB"
                else:
                    size_str = str((size + 1024**2 // 2) // 1024**2) + "  MB"

                self.path_listbox.insert(tk.END, fp)
                self.format_listbox.insert(tk.END, BITMAP_FORMATS[info.format])
                self.type_listbox.insert(tk.END, BITMAP_TYPES[info.type])
                self.size_listbox.insert(tk.END, size_str)

                self.update_path_listbox_entry_color(tk.END)

            self.synchronize_selection()
        except Exception:
            print(format_exc())

        self._populating = False

    def update_path_listbox_entry_color(self, i):
        fp = self.path_listbox.get(i)
        if self.master.get_will_be_processed(fp):
            self.path_listbox.itemconfig(i, bg='dark green', fg='white')
        else:
            self.path_listbox.itemconfig(i, bg=self.enum_normal_color,
                                         fg=self.text_normal_color,)

    def _scroll_all_yviews(self, *args):
        for listbox in self.listboxes:
            listbox.yview(*args)

    def _sync_yviews(self, src_listbox, *args):
        for listbox in self.listboxes:
            if listbox is not src_listbox and src_listbox.yview() != listbox.yview():
                listbox.yview_moveto(args[0])
        self.vsb.set(*args)

    def _path_scrolled(self, *args):
        self._sync_yviews(self.path_listbox, *args)

    def _format_scrolled(self, *args):
        self._sync_yviews(self.format_listbox, *args)

    def _type_scrolled(self, *args):
        self._sync_yviews(self.type_listbox, *args)

    def _size_scrolled(self, *args):
        self._sync_yviews(self.size_listbox, *args)


if __name__ == "__main__":
    try:
        BitmapConverterWindow(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#!/usr/bin/env python3
#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

try:
    from .converter_base import ConverterBase
except (ImportError, SystemError):
    from converter_base import ConverterBase

from pathlib import Path
import threadsafe_tkinter as tk

from traceback import format_exc
from struct import pack_into, pack

from binilla.windows.filedialog import askopenfilename

from reclaimer.hek.defs.coll import coll_def
from reclaimer.hek.defs.mod2 import mod2_def
from reclaimer.hek.defs.mode import mode_def
from reclaimer.stubbs.defs.mode import mode_def as stubbs_mode_def
from reclaimer.stubbs.defs.coll import coll_def as stubbs_coll_def
from reclaimer.common_descs import tag_header
from reclaimer.util.matrices import quaternion_to_matrix, Matrix

from supyr_struct.buffer import get_rawdata
from supyr_struct.defs.block_def import BlockDef
from supyr_struct.field_types import FieldType, BytearrayRaw

window_base_class = tk.Toplevel
if __name__ == "__main__":
    window_base_class = tk.Tk

tag_header_def = BlockDef(tag_header)


def undef_size(node, *a, **kwa):
    if node is None:
        return 0
    return len(node)


raw_block_def = BlockDef("raw_block",
    BytearrayRaw('data', SIZE=undef_size)
    )


class Permutation():
    name = ""
    nodes = None
    def __init__(self, name):
        self.name = name
        self.nodes = {}


class Region():
    name = ""
    perms = None
    def __init__(self, name):
        self.name = name
        self.perms = {}


def get_tags(coll_path, model_in_path):
    mod2_path = str(Path(model_in_path).with_suffix('')) + "_COLL.gbxmodel"

    # get whether or not the collision tag is stubbs
    stubbs = tag_header_def.build(filepath=coll_path).version == 11

    if stubbs:
        coll_tag = stubbs_coll_def.build(filepath=coll_path)
    else:
        coll_tag = coll_def.build(filepath=coll_path)

    mod2_tag = mod2_def.build()
    mod2_tag.filepath = mod2_path
    model_in_rawdata = None

    guessed_mode = False
    while model_in_rawdata is None and model_in_path:
        try:
            model_in_rawdata = get_rawdata(filepath=model_in_path)
        except Exception:
            if guessed_mode:
                model_in_rawdata = None
                model_in_path = Path(askopenfilename(
                    initialdir=model_in_path.parent, filetypes=(
                        ('All', '*'), ('Gbxmodel', '*.gbxmodel')),
                    title="Select the gbxmodel to extract nodes from"))
            else:
                model_in_path = model_in_path.with_suffix(".model")
                guessed_mode = True

    if model_in_rawdata is not None:
        # we dont actually care about the geometries or shaders of the gbxmodel
        # tag we're loading, so null them out to speed up the loading process.
        geom_off = 64 + 4*9 + 2*5 + 126 + 12*3

        # make a copy so we dont edit the file
        model_in_rawdata = bytearray(model_in_rawdata)
        model_in_rawdata[geom_off:64 + 232] = b'\x00'*(64 + 232 - geom_off)

        if model_in_rawdata[36:40] == b"mod2":
            model_in_tag = mod2_def.build(rawdata=model_in_rawdata)
        elif stubbs:
            model_in_tag = stubbs_mode_def.build(rawdata=model_in_rawdata)
        else:
            model_in_tag = mode_def.build(rawdata=model_in_rawdata)

        mod2_tag.data.tagdata.nodes = model_in_tag.data.tagdata.nodes
    else:
        model_in_tag = None
        mod2_tag.data.tagdata.nodes.STEPTREE.append()
        node = mod2_tag.data.tagdata.nodes.STEPTREE[-1]
        node.name = "COLLISION ROOT"
        print("    %s" % model_in_path)
        print("    Could not load gbxmodel. Gbxmodel wont have nodes and "
              "the geometry will not be positioned or rotated properly.")

    return coll_tag, model_in_tag, mod2_tag


def get_collections(coll_tag):
    node_names = []
    reg_names = []
    all_perm_names = {}
    sorted_bsps = {}

    # build the region and permutation collections
    for region in coll_tag.data.tagdata.regions.STEPTREE:
        perms_by_region = Region(region.name)
        reg_names.append(region.name)

        perm_names = []
        all_perm_names[region.name] = perm_names

        for perm in region.permutations.STEPTREE:
            node_bsps_by_perm = Permutation(perm.name)
            perm_names.append(perm.name)

            perms_by_region.perms[perm.name] = node_bsps_by_perm

        sorted_bsps[region.name] = perms_by_region

    for node in coll_tag.data.tagdata.nodes.STEPTREE:
        node_names.append(node.name)

    return sorted_bsps, reg_names, all_perm_names, node_names


def get_node_transforms(nodes):
    node_transforms = [None]*len(nodes)

    # create a collection of absolute translations and rotations
    # to move and rotate each nodes collision model into place.
    for node_i in range(len(nodes)):
        node = nodes[node_i]
        trans = node.translation
        rot = node.rotation
        x, y, z = trans.x, trans.y, trans.z
        this_rot = quaternion_to_matrix(rot.i, rot.j, rot.k, rot.w)

        if node.parent_node >= 0:
            parent = node_transforms[node.parent_node]
            trans = Matrix((x, y, z))

            parent_rot = parent[4]
            this_rot = parent_rot * this_rot

            trans = parent_rot * trans
            x = trans[0][0] + parent[0]
            y = trans[1][0] + parent[1]
            z = trans[2][0] + parent[2]
        else:
            parent_rot = quaternion_to_matrix(0, 0, 0, 1)

        node_transforms[node_i] = [x, y, z, parent_rot, this_rot]

    return node_transforms


def sort_node_bsps(coll_tag, sorted_bsps, reg_names, all_perm_names):
    node_bsps = coll_tag.data.tagdata.nodes.STEPTREE
    coll_data = coll_tag.data.tagdata

    # sort the bsps by region, permutation, and node
    for node_i in range(len(node_bsps)):
        node = node_bsps[node_i]
        node_name = node.name

        try:
            region_name = reg_names[node.region]
            perm_names = all_perm_names[region_name]
            region = coll_data.regions.STEPTREE[node.region]
        except (IndexError, KeyError):
            continue

        bsps = node.bsps.STEPTREE
        perms_by_region = sorted_bsps[region_name].perms

        for bsp_i in range(len(bsps)):
            bsp = bsps[bsp_i]
            try:
                node_bsps_by_perm = perms_by_region[perm_names[bsp_i]]
            except (IndexError, KeyError):
                continue

            node_bsps_by_perm.nodes[node_name] = bsp

def make_raw_verts_block(node_bsp, node_transforms, node_i):
    raw_verts = raw_block_def.build()

    # create the uncompressed vertices bytearray that
    # all the parts will reuse. This is wasteful, but
    # i dont care to convert the vertex indices, so w/e
    uncomp_verts = bytearray(68*len(node_bsp.vertices.STEPTREE))
    raw_verts.data = uncomp_verts

    transform = node_transforms[node_i]
    dx, dy, dz = transform[0], transform[1], transform[2]
    rotation = transform[4]
    pos = 0
    for vert in node_bsp.vertices.STEPTREE:
        # rotate the vertices to match the nodes orientation
        trans = rotation * Matrix(vert[:3])

        # also translate the vertices to match the nodes position
        pack_into('>14f2h2f', uncomp_verts, pos,
                  trans[0][0]+dx, trans[1][0]+dy, trans[2][0]+dz,
                  1,0,0,0,1,0,0,0,1,0,0,0,0,1,0)
        pos += 68

    return raw_verts


def make_parts_by_mats(faces, raw_verts, parts, shaders, node_i,
                       use_mats=True, has_frames=True):
    parts_by_mats = {}
    for face in faces:
        mat = face.material
        missing_shader_ct = (mat + 1) - len(shaders)
        if missing_shader_ct > 0:
            for i in range(missing_shader_ct):
                shaders.append()
                shaders[-1].shader.filepath = "MISSING"

        if mat in parts_by_mats:
            continue
        if use_mats or not parts_by_mats:
            parts.append()
        parts_by_mats[mat] = part = parts[-1]

        part.shader_index = mat
        part.local_node_count = 1

        # if there are no frames, all parts
        # will be left parented to frame 0
        if has_frames:
            part.local_nodes[0] = node_i

        part.uncompressed_vertices.size = len(raw_verts.data)//68
        part.uncompressed_vertices.STEPTREE = raw_verts

    return parts_by_mats


def get_edge_loops_by_mats(faces, edges):
    edge_loops_by_mats = {}
    for f_i in range(len(faces)):
        face = faces[f_i]
        mat = face.material
        if mat in edge_loops_by_mats:
            edge_loops = edge_loops_by_mats[mat]
        else:
            edge_loops_by_mats[mat] = edge_loops = []

        e_i = face.first_edge
        face_edges = set()
        vert_indices = []
        while e_i not in face_edges:
            face_edges.add(e_i)
            edge = edges[e_i]
            if edge[4] == f_i:
                e_i = edge[2]
                vert_indices.append(edge[0])
            else:
                e_i = edge[3]
                vert_indices.append(edge[1])

        edge_loops.append(vert_indices)

    return edge_loops_by_mats


def fill_parts_by_mats(parts_by_mats, edge_loops_by_mats):
    for mat in edge_loops_by_mats:
        loops = edge_loops_by_mats[mat]
        part = parts_by_mats[mat]

        # create the uncompressed triangles bytearray
        tris = part.triangles
        tris.STEPTREE = raw_block_def.build()

        uncomp_tris = b''
        rev = True
        for loop in loops:
            v0 = loop[0]
            v1 = loop[1]
            v0_packed = pack('>h', v0)
            if uncomp_tris:
                uncomp_tris += v0_packed
                rev = not rev

            for v2 in loop[2:]:
                if rev:
                    uncomp_tris += v0_packed*2 + pack('>hh', v1, v2)
                else:
                    uncomp_tris += v0_packed*2 + pack('>hh', v2, v1)

                rev = not rev

                v1 = v2

                # repeat the last vert to break the strip
                uncomp_tris += uncomp_tris[-2:]

        # cut off the last repeated verts
        uncomp_tris = uncomp_tris[:-2]

        # pad the last strip piece up to 3 entries
        uncomp_tris += uncomp_tris[-2:]*(
            (3-((len(uncomp_tris)//2)%3))%3)

        tris.size = len(uncomp_tris)//6
        tris.STEPTREE.data = uncomp_tris


def coll_to_mod2(coll_path, model_in_path=None, guess_mod2=True, use_mats=True):
    if guess_mod2:
        model_in_path = Path(coll_path).with_suffix(".gbxmodel")

    print("    Loading tags")
    coll_tag, model_in_tag, mod2_tag = get_tags(coll_path, model_in_path)
    mod2_data = mod2_tag.data.tagdata


    # TODO: MAKE SURE REGIONS AND PERMUTATIONS MATCH BETWEEN model_in_tag AND coll_tag
    mod2_data.flags.parts_have_local_nodes = True
    mod2_data.base_map_u_scale = 1.0
    mod2_data.base_map_v_scale = 1.0

    nodes   = mod2_data.nodes.STEPTREE
    regions = mod2_data.regions.STEPTREE
    geoms   = mod2_data.geometries.STEPTREE
    shaders = mod2_data.shaders.STEPTREE

    #print("Making materials")
    # build the shaders to represent the materials
    for material in coll_tag.data.tagdata.materials.STEPTREE:
        shaders.append()
        shaders[-1].shader.filepath = material.name

    #print("Getting absolute transforms")
    node_transforms = get_node_transforms(nodes)

    #print("Making collections")
    sorted_bsps, reg_names, perm_names, node_names = get_collections(coll_tag)

    #print("Sorting bsps")
    sort_node_bsps(coll_tag, sorted_bsps, reg_names, perm_names)


    print("    Building regions, permutations, geometries, and parts")
    # build the regions and permutations blocks
    for r_name in reg_names:
        perms_by_region = sorted_bsps[r_name].perms
        first_geom = len(geoms)
        perm_count = len(perms_by_region)

        # create a new region
        regions.append()
        region = regions[-1]

        region.name = r_name
        perms = region.permutations.STEPTREE

        # create as many perms and geoms
        # as we'll need for this region
        for p_name in perm_names[r_name]:
            # make a new geometry to hold all the parts we'll be making
            geoms.append()
            geom = geoms[-1]
            parts = geom.parts.STEPTREE
            node_bsps_by_perm = perms_by_region[p_name].nodes

            # create as many parts as we need for this geometry
            for node_i in range(len(node_names)):
                node_name = node_names[node_i]
                if not model_in_tag:
                    node_i = 0

                if node_name not in node_bsps_by_perm:
                    continue

                node_bsp = node_bsps_by_perm[node_name]
                faces = node_bsp.surfaces.STEPTREE
                edges = node_bsp.edges.STEPTREE

                raw_verts = make_raw_verts_block(
                    node_bsp, node_transforms, node_i)

                # find out how many materials are in this node_bsp
                # and create parts blocks for each of them.
                parts_by_mats = make_parts_by_mats(
                    faces, raw_verts, parts, shaders,
                    node_i, use_mats, model_in_tag)

                # collect all the edge loops that make up all the faces
                edge_loops_by_mats = get_edge_loops_by_mats(faces, edges)

                # loop over each material and create the
                # triangle list for the edges loops in it.
                fill_parts_by_mats(parts_by_mats, edge_loops_by_mats)

            # if we didnt make any parts, this geometry
            # is actually empty and should be deleted.
            if not len(parts):
                del geoms[-1]
                continue

            # create a permutation for this geometry
            perms.append()
            perm = perms[-1]
            perm.name = p_name
            geom_index = len(geoms) - 1
            perm.superlow_geometry_block = geom_index
            perm.low_geometry_block = geom_index
            perm.medium_geometry_block = geom_index
            perm.high_geometry_block = geom_index
            perm.superhigh_geometry_block = geom_index

    return mod2_tag


class CollisionConverter(ConverterBase, window_base_class):
    src_ext = "model_collision_geometry"
    dst_ext = "gbxmodel"

    def __init__(self, app_root, *args, **kwargs):
        if isinstance(self, tk.Toplevel):
            kwargs.update(bd=0, highlightthickness=0, bg=self.default_bg_color)

        window_base_class.__init__(self, app_root, *args, **kwargs)
        ConverterBase.__init__(self, app_root, *args, **kwargs)
        self.setup_window(*args, **kwargs)

    def setup_window(self, *args, **kwargs):
        ConverterBase.setup_window(self, *args, **kwargs)

        self.guess_mod2 = tk.IntVar(self, 1)
        self.use_mats = tk.IntVar(self, 1)

        # make the frames
        self.settings_frame = tk.LabelFrame(self, text="Conversion settings")

        # add the buttons
        self.guess_mod2_checkbutton = tk.Checkbutton(
            self.settings_frame, variable=self.guess_mod2,
            text="Locate gbxmodel in directory")
        self.use_mats_checkbutton = tk.Checkbutton(
            self.settings_frame, variable=self.use_mats,
            text="Use collision materials as shaders")

        self.pack_widgets()
        self.apply_style()

    def pack_widgets(self):
        ConverterBase.pack_widgets(self)

        # pack everything
        self.guess_mod2_checkbutton.pack(anchor='w', padx=10)
        self.use_mats_checkbutton.pack(anchor='w', padx=10)
        self.settings_frame.pack(fill='both', anchor='nw')

    def destroy(self):
        ConverterBase.destroy(self)
        window_base_class.destroy(self)

    def convert(self, tag_path):
        return coll_to_mod2(tag_path, guess_mod2=self.guess_mod2.get(),
                            use_mats=self.use_mats.get())

    def lock_ui(self):
        ConverterBase.lock_ui(self)
        for w in (self.guess_mod2_checkbutton, self.use_mats_checkbutton):
            w.config(state=tk.DISABLED)

    def unlock_ui(self):
        ConverterBase.unlock_ui(self)
        for w in (self.guess_mod2_checkbutton, self.use_mats_checkbutton):
            w.config(state=tk.NORMAL)


if __name__ == "__main__":
    try:
        CollisionConverter(None).mainloop()
        raise SystemExit(0)
    except Exception:
        print(format_exc())
        input()



#
# This file is part of Mozzarilla.
#
# For authors and copyright check AUTHORS.TXT
#
# Mozzarilla is free software under the GNU General Public License v3.0.
# See LICENSE for more information.
#

import os

from pathlib import Path
from traceback import format_exc

from reclaimer.model.jms import read_jms
from reclaimer.physics.physics_compilation import compile_physics
from supyr_struct.util import is_path_empty
from binilla.windows.filedialog import askopenfilename


def physics_from_jms(app, fp=None):
    load_dir = app.jms_load_dir
    tags_dir = app.tags_dir
    data_dir = app.data_dir
    if is_path_empty(tags_dir):
        tags_dir = Path("")
    if is_path_empty(data_dir):
        data_dir = Path("")

    if is_path_empty(load_dir):
        load_dir = data_dir

    if is_path_empty(fp):
        fp = askopenfilename(
            initialdir=load_dir, parent=app,
            filetypes=(("JMS model", "*.jms"), ("All", "*")),
            title="Select jms file to turn into a physics tag")

    fp = Path(fp)
    if is_path_empty(fp):
        return

    try:
        app.jms_load_dir = fp.parent
        with fp.open("r") as f:
            jms_model = read_jms(f.read(), "regions")
    except Exception:
        print(format_exc())
        print("    Could not parse jms file")
        return

    try:
        rel_filepath = fp.relative_to(data_dir).parent.parent
        rel_filepath = rel_filepath.joinpath(rel_filepath.stem + ".physics")
    except ValueError:
        rel_filepath = Path("unnamed.physics")

    tag_path = Path("")
    if not is_path_empty(rel_filepath):
        tag_path = tags_dir.joinpath(rel_filepath)

    # make the tag window
    window = app.load_tags(
        filepaths=(tag_path, ) if tag_path.is_file() else "",
        def_id='phys')
    if not window:
        return

    window = window[0]
    window.is_new_tag = False
    window.tag.filepath = tag_path
    window.tag.rel_filepath = rel_filepath

    compile_physics(window.tag, jms_model.markers, tag_path.is_file())

    # reload the window to display the newly entered info
    window.reload()
    app.update_tag_window_title(window)
    if not tag_path.is_file():
        # save the tag if it doesnt already exist
        app.save_tag()
